import requests
from tqdm import tqdm
import time
from bs4 import BeautifulSoup
import numpy as np


def get_url_from_Mastr_website() -> str:
    """Get the url of the latest MaStR file from markstammdatenregister.de.

    The file and the corresponding url are updated once per day.
    The url has a randomly generated string appended, so it has to be
    grabbed from the marktstammdatenregister.de homepage.
    For further details visit https://www.marktstammdatenregister.de/MaStR/Datendownload
    """

    html = requests.get("https://www.marktstammdatenregister.de/MaStR/Datendownload")
    soup = BeautifulSoup(html.text, "lxml")
    # find the download button element on the website
    element = soup.find_all("a", "btn btn-primary text-right")[0]
    # extract the url from the html element
    url = str(element).split('href="')[1].split('" title')[0]
    return url


def download_xml_Mastr(save_path: str) -> None:
    """Downloads the zipped MaStR.

    Parameters
    -----------
    save_path: str
        The path where the downloaded MaStR zipped folder will be saved.
    """
    print_message = (
        "Download has started, this can take several minutes."
        "The download bar is only a rough estimate."
    )
    print(print_message)
    url = get_url_from_Mastr_website()
    time_a = time.perf_counter()
    r = requests.get(url, stream=True)
    total_length = int(10000 * 1024 * 1024)
    with open(save_path, "wb") as zfile, tqdm(
        desc=save_path, total=(total_length / 1024 / 1024)
    ) as bar:
        for chunk in r.iter_content(chunk_size=1024 * 1024):
            # chunk size of 1024 * 1024 needs 9min 11 sek = 551sek
            # chunk size of 1024 needs 9min 11 sek as well
            if chunk:
                zfile.write(chunk)
                zfile.flush()
            bar.update()
    time_b = time.perf_counter()
    print(f"Download is finished. It took {int(np.around(time_b - time_a))} seconds.")



from open_mastr.soap_api.config import setup_project_home

setup_project_home()

# This import should be called after the setup project home, since .open_mastr folder must be defined first
from .mastr import Mastr  # noqa: E402 ignore import order warning of flake8



from contextlib import contextmanager
from sqlalchemy import create_engine
from sqlalchemy.orm import Query, sessionmaker

from open_mastr.settings import DB_URL


def chunks(lst, n):
    """Yield successive n-sized chunks from lst.

    `Credits <https://stackoverflow.com/questions/312443/how-do-you-split-a-list-into-evenly-sized-chunks>`_
    """
    if isinstance(lst, Query):
        length = lst.count()
    else:
        length = len(lst)
    for i in range(0, length, n):
        yield lst[i: i + n]


def db_engine(): # TODO: Include in _create_database in MaStR() class
    return create_engine(DB_URL)


@contextmanager
def session_scope():
    """Provide a transactional scope around a series of operations."""
    Session = sessionmaker(bind=db_engine())
    session = Session()
    try:
        yield session
        session.commit()
    except:
        session.rollback()
        raise
    finally:
        session.close()



import datetime
import multiprocessing
import logging
import tqdm

from open_mastr.soap_api.utils import is_time_blacklisted
log = logging.getLogger(__name__)

last_successful_download = datetime.datetime.now()

# Check if worker threads need to be killed


def _stop_execution(time_blacklist, timeout):
    # Last successful execution was more than 10 minutes ago. Server seems
    # unresponsive, so stop execution permanently by raising an error
    if last_successful_download + datetime.timedelta(minutes=timeout) < datetime.datetime.now():
        log.error('No response from server in the last {} minutes. Stopping execution.'.format(timeout))
        raise ConnectionAbortedError
    # Stop execution smoothly if current system time is in blacklist by
    # returning. Calling function can decide to continue running later if
    # needed
    if time_blacklist and is_time_blacklisted(last_successful_download.time()):
        log.info('Current time is in blacklist. Halting.')
        return True
    # ... Add more checks here if needed ...
    return False



import csv
import datetime
import json
import os
import uuid

from open_mastr.soap_api.metadata.description import DataDescription
from open_mastr.soap_api.config import get_data_config, get_filenames, column_renaming
from open_mastr.soap_api.download import MaStRDownload


# TODO: We should not describe the data in both metadata folder and orm.py


def datapackag_base(reference_date, publication_date=None, statistik_flag=None):
    """
    Create datapackage metadata base information

    Includes all information about the data except for key `resources` which describes the actual data files and its
    columns.

    Parameters
    ----------
    reference_date: datetime.datetime
        Reference date for data
    publication_date: datetime.datetime
        Date of data being published. Defaults to `datetime.datetime.now()`, the time when this metadata is being
        created.
    statistik_flag: str or None
        Describe if filtering is applied during CSV export of data. Read in
        :meth:`~.open_mastr.soap_api.mirror.MaStRMirror.to_csv()` for more details.

    Returns
    -------
    dict
        Datapackage metadata
    """

    if not publication_date:
        publication_date = datetime.datetime.now()

    publication_date = publication_date.strftime("%Y-%m-%d %H:%M:%S")
    data_version = get_data_config()["data_version"]

    # Add a note to the description if this is filtered data or if it the complete data including potential duplicates
    if statistik_flag == "B":
        description_extra = (
            "The original MaStR data is filtered by `StatistikFlag == 'B'`. It includes data that "
            "was migrated to the Martstammdatenregister + newly registered units with commissioning "
            "date after 31.01.2019 Thus, it is suitable for "
            "statistical analysis as is does not contain duplicates.\n"
        )
    elif statistik_flag == "A":
        description_extra = (
            "The original MaStR data is filtered by `StatistikFlag == 'A'`. "
            "It includes newly registered units with commissioning date before 31.01.2019. "
            "Data may contain duplicates.\n"
        )
    else:
        description_extra = "All data from the Marktstammdatenregister is included. There are duplicates included.\n"
    description_extra += (
        "For further information read in the documentation of the original data source: "
        "https://www.marktstammdatenregister.de/MaStRHilfe/subpages/statistik.html"
    )

    datapackage_meta = {
        "name": "open-mastr_raw",
        "title": "open-MaStR power unit registry",
        "id": str(uuid.uuid4()),
        "description": f"Raw data download Marktstammdatenregister (MaStR) data using the webservice.\n\n{description_extra}",
        "language": ["en-GB", "de-DE"],
        "keywords": ["powerplants", "renewables"],
        "created": publication_date,
        "version": data_version,
        "context": {
            "homepage": "https://www.marktstammdatenregister.de/MaStR/",
            "documentation": "https://www.marktstammdatenregister.de/MaStRHilfe/index.html",
            "sourceCode": None,
            "contact": "https://www.marktstammdatenregister.de/MaStR/Startseite/Kontakt",
            "grantNo": None,
            "fundingAgency": "Bundesnetzagentur für Elektrizität, Gas, Telekommunikation, Post und Eisenbahnen",
            "fundingAgencyLogo": "https://www.bundesnetzagentur.de/DE/Home/home_node.html",
            "publisherLogo": "https://www.bundesnetzagentur.de/DE/Sachgebiete/ElektrizitaetundGas/Unternehmen_Institutionen/MaStR/MaStR_logo_long.svg?__blob=normal&v=4",
        },
        "spatial": {"location": None, "extent": "Germany", "resolution": "vector"},
        "temporal": {
            "referenceDate": reference_date.strftime("%Y-%m-%d %H:%M:%S"),
            "timeseries": {
                "start": None,
                "end": None,
                "resolution": None,
                "alignment": None,
                "aggregationType": None,
            },
        },
        "sources": [
            {
                "title": "Bundesnetzagentur - Marktstammdatenregister",
                "description": "Das Marktstammdatenregister ist das Register für den deutschen Strom- und Gasmarkt. Es wird MaStR abgekürzt. Im MaStR sind vor allem die Stammdaten zu Strom- und Gaserzeugungsanlagen zu registrieren. Außerdem sind die Stammdaten von Marktakteuren wie Anlagenbetreibern, Netzbetreibern und Energielieferanten zu registrieren. Das MaStR wird von der Bundesnetzagentur geführt.",
                "path": "https://www.marktstammdatenregister.de/MaStR/",
                "licenses": [
                    {
                        "name": "dl-de/by-2-0",
                        "title": "Datenlizenz Deutschland – Namensnennung – Version 2.0",
                        "path": "https://www.govdata.de/dl-de/by-2-0",
                        "instruction": "You are free: To Share, To Create, To Adapt; As long as you: Attribute",
                        "attribution": f"© Marktstammdatenregister {datetime.date.today().year} | dl-de/by-2-0",
                    }
                ],
            },
            {
                "title": "RLI - open_MaStR",
                "description": "Scripts to download, process and publish the MaStR data set.",
                "path": "https://github.com/OpenEnergyPlatform/open-MaStR",
                "licenses": [
                    {
                        "name": "AGPL-3.0",
                        "title": "GNU Affero General Public License v3.0",
                        "path": "https://www.gnu.org/licenses/agpl-3.0.en.html",
                        "instruction": "You are free: To Share, To Create, To Adapt; As long as you: Attribute, Share-Alike, Keep open!",
                        "attribution": "open_MaStR © Reiner Lemoine Institut | AGPL-3.0",
                    }
                ],
            },
        ],
        "licenses": [
            {
                "name": "dl-de/by-2-0",
                "title": "Datenlizenz Deutschland – Namensnennung – Version 2.0",
                "path": "https://www.govdata.de/dl-de/by-2-0",
                "instruction": "Die bereitgestellten Daten und Metadaten dürfen für die kommerzielle und nicht kommerzielle Nutzung verwendet werden. Bei der Nutzung ist sicherzustellen, dass Angaben als Quellenvermerk enthalten sind.",
                "attribution": "© Marktstammdatenregister 2019 | dl-de/by-2-0",
            }
        ],
        "contributors": [
            {
                "title": "Ludee",
                "email": None,
                "path": "https://github.com/ludee",
                "role": "maintainer",
                "organization": "Reiner Lemoine Institut gGmbH",
            },
            {
                "title": "Guido Pleßmann",
                "email": None,
                "path": "https://gplssm.de",
                "role": "maintainer",
                "organization": "Reiner Lemoine Institut gGmbH",
            },
            {
                "title": "oakca",
                "email": None,
                "path": "https://github.com/oakca",
                "role": "contributor",
                "organization": "Reiner Lemoine Institut gGmbH",
            },
        ],
        "review": {"path": None, "badge": None},
        "metaMetadata": {
            "metadataVersion": "OEP-1.4.0",
            "metadataLicense": {
                "name": "CC0-1.0",
                "title": "Creative Commons Zero v1.0 Universal",
                "path": "https://creativecommons.org/publicdomain/zero/1.0/",
            },
        },
        "_comment": {
            "metadata": "Metadata documentation and explanation (https://github.com/OpenEnergyPlatform/organisation/wiki/metadata)",
            "dates": "Dates and time must follow the ISO8601 including time zone (YYYY-MM-DD or YYYY-MM-DDThh:mm:ss±hh)",
            "units": "Use a space between numbers and units (100 m)",
            "languages": "Languages must follow the IETF (BCP47) format (en-GB, en-US, de-DE)",
            "licenses": "License name must follow the SPDX License List (https://spdx.org/licenses/)",
            "review": "Following the OEP Data Review (https://github.com/OpenEnergyPlatform/data-preprocessing/wiki)",
            "null": "If not applicable use (null)",
        },
    }

    return datapackage_meta


def datapackage_meta_json(
    reference_date,
    technologies=None,
    data=["raw"],
    statistik_flag=None,
    json_serialize=True,
):
    """
    Create a frictionless data conform metadata description

    Parameters
    ----------
    reference_date: datetime.datetime
        Reference date for data
    technologies: list
        Only consider specified technologies in metadata JSON string. If not provided, metadata is created for data
        of all technologies.
    data: list, optional
        Specify which data should be documented by this metadata.
    statistik_flag: str or None
        Describe if filtering is applied during CSV export of data. Read in
        :meth:`~.open_mastr.soap_api.mirror.MaStRMirror.to_csv()` for more details.
    json_serialize: bool
        Toggle between returning a JSON string or a dict

        * True: return JSON string
        * False: return dict

    Returns
    -------
    dict or str
        Set `json_serialize` to False, for returning a dict instead of JSON str.
    """

    datapackage_base_dict = datapackag_base(
        reference_date, statistik_flag=statistik_flag
    )

    table_columns = DataDescription().functions_data_documentation()
    mastr_dl = MaStRDownload()

    # Filter specified technologies
    if technologies:
        unit_data_specs = {
            k: v for k, v in mastr_dl._unit_data_specs.items() if k in technologies
        }

    filenames = get_filenames()

    renaming = column_renaming()

    resources_meta = {"resources": []}

    # Add resources of raw data files
    for tech, specs in unit_data_specs.items():
        raw_fields = []
        specs["basic_data"] = "GetListeAlleEinheiten"

        for data_type in [
            "basic_data",
            "unit_data",
            "eeg_data",
            "kwk_data",
            "permit_data",
        ]:
            if data_type in specs.keys():
                for name, specification in table_columns[specs[data_type]].items():
                    if name in renaming[data_type]["columns"]:
                        name = f"{name}_{renaming[data_type]['suffix']}"
                    raw_fields.append({"name": name, **specification, "unit": None})

        if "raw" in data:
            resource = {
                "profile": "tabular-data-resource",
                "name": f"bnetza_mastr_{tech}_raw",
                "title": f"open-MaStR {tech} units (raw)",
                "path": filenames["raw"][tech]["joined"],
                "scheme": "file",
                "encoding": "utf-8",
                "mediatype": "text/csv",
                "schema": {
                    "fields": raw_fields,
                    "primaryKey": ["EinheitMastrNummer"],
                },
                "dialect": {"delimiter": ","},
            }

            resources_meta["resources"].append(resource)
        if "cleaned" in data:
            resource = {
                "profile": "tabular-data-resource",
                "name": f"bnetza_mastr_{tech}_cleaned",
                "title": f"open-MaStR {tech} units (cleaned)",
                "path": filenames["cleaned"][tech],
                "scheme": "file",
                "encoding": "utf-8",
                "mediatype": "text/csv",
                "schema": {
                    "fields": raw_fields,
                    "primaryKey": ["EinheitMastrNummer"],
                },
                "dialect": {"delimiter": ","},
            }

            resources_meta["resources"].append(resource)
        if "postprocessed" in data:
            processed_fields = [
                {
                    "name": "geom",
                    "unit": None,
                    "type": "str",
                    "desciption": "Standort der Anlage als Punktgeometrie im WKB Format",
                    "examples": "0101000020e610000071fbe59315131c40a2b437f8c20e4a40",
                },
                {
                    "name": "comment",
                    "unit": None,
                    "type": "str",
                    "desciption": "Information about data post-processing",
                    "examples": "has_geom; outside_vg250",
                },
            ]
            if tech == "wind":
                processed_fields.append(
                    {
                        "name": "tags",
                        "unit": None,
                        "type": "json",
                        "desciption": "Data insights and report about post-processing steps",
                        "examples": {
                            "plz_check": False,
                            "processed": True,
                            "inside_germany": True,
                        },
                    }
                )
                processed_fields.append(
                    {
                        "name": "geom",
                        "unit": None,
                        "type": "str",
                        "desciption": "Standort der Anlage als Punktgeometrie im WKB Format (EPSG 3035)",
                        "examples": "0101000020e610000071fbe59315131c40a2b437f8c20e4a40",
                    }
                )
            resource = {
                "profile": "tabular-data-resource",
                "name": f"bnetza_mastr_{tech}",
                "title": f"open-MaStR {tech} units",
                "path": filenames["postprocessed"][tech],
                "scheme": "file",
                "encoding": "utf-8",
                "mediatype": "text/csv",
                "schema": {
                    "fields": raw_fields + processed_fields,
                    "primaryKey": ["EinheitMastrNummer"],
                },
                "dialect": {"delimiter": ","},
            }

            resources_meta["resources"].append(resource)

    datapackage_dict = {**datapackage_base_dict, **resources_meta}

    if json_serialize:
        return json.dumps(datapackage_dict, ensure_ascii=False)
    else:
        return datapackage_dict


def column_docs_csv(technologies, base_path):
    metadata = datapackage_meta_json(
        datetime.datetime.now(), technologies, json_serialize=False
    )

    filenames = []

    # Sort filenames according to technologies
    raw_data_filenames = get_filenames()["raw"]
    metadata_resources = []

    for tech in technologies:
        for metadata_resource in metadata["resources"]:
            if metadata_resource["path"] == raw_data_filenames[tech]["joined"]:
                metadata_resources.append(metadata_resource)

    for table in metadata_resources:
        csv_header = ["Column", "Description", "Type", "Example"]

        os.makedirs(base_path, exist_ok=True)
        filename_base = table["name"] + ".csv"
        filename = os.path.join(base_path, filename_base)
        filenames.append(filename_base)
        with open(filename, "w", encoding="utf-8", newline="") as fh:
            fh_writer = csv.writer(fh)
            fh_writer.writerow(csv_header)
            for column in table["schema"]["fields"]:
                csv_row = [
                    column["name"],
                    column["description"],
                    column["type"],
                    column["example"],
                ]
                fh_writer.writerow(csv_row)

    return filenames



import os
from os.path import expanduser

# TODO: Migrate the DB settings from the bulk download
DB_ENGINE = os.environ.get("DB_ENGINE", "sqlite")

SQLITE_DATABASE_PATH = os.environ.get("SQLITE_DATABASE_PATH", os.path.join(
            expanduser("~"), ".open-MaStR", "data", "sqlite", "open-mastr.db"
        ))

DB_URL = (
    "postgresql+psycopg2://open-mastr:open-mastr@localhost:55443/open-mastr"
    if DB_ENGINE == "docker" else f"sqlite:///{SQLITE_DATABASE_PATH}"
)



__copyright__ = "© Reiner Lemoine Institut"
__license__ = "GNU Affero General Public License Version 3 (AGPL-3.0)"
__url__ = "https://www.gnu.org/licenses/agpl-3.0.en.html"
__author__ = "Ludee; Bachibouzouk; solar-c"
__issue__ = "https://github.com/OpenEnergyPlatform/examples/issues/52"
__version__ = "v0.10.0"


import os
import pandas as pd
import logging
import datetime


log = logging.getLogger(__name__)


def is_time_blacklisted(time):
    times_blacklist = [
        ('8:00', '18:00'),  # BNetzA Business hours
        ('23:30', '00:10'),  # Daily database cronjob
        # Add more if needed...
    ]

    # check if time is in a given interval between upper and lower
    def in_interval(lower, upper):
        # Convert str to datatime object
        def parse_time(t): 
            return datetime.datetime.strptime(t, "%H:%M").time()
        lower = parse_time(lower)
        upper = parse_time(upper)

        # Handle interval that spans over midnight (i.e. 23:30-0:30)
        if lower > upper:
            return (time <= upper or time >= lower)
        # Handle all other intevals
        return (lower <= time and upper >= time)

    # check if time is in interval for each interval in the blacklist
    in_interval = [in_interval(lower, upper) for lower, upper in times_blacklist]
    return any(in_interval)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Service functions for logging

Configure console and file logging; Create and handle config file for api keys;

SPDX-License-Identifier: AGPL-3.0-or-later
"""

__copyright__ = "© Reiner Lemoine Institut"
__license__ = "GNU Affero General Public License Version 3 (AGPL-3.0)"
__url__ = "https://www.gnu.org/licenses/agpl-3.0.en.html"
__author__ = "Ludee; christian-rli"
__issue__ = "https://github.com/OpenEnergyPlatform/examples/issues/52"
__version__ = "v0.10.0"

import os
import yaml
import shutil
import pathlib

import logging
import logging.config
log = logging.getLogger(__name__)


def get_project_home_dir():
    """Get root dir of project data
    
    On linux this path equals `$HOME/open-MaStR/`, respectively `~/open-MaStR/` 
    which is also called `PROJECTHOME`.
    
    Returns
    -------
    path-like object
        Absolute path to root dir of open-MaStR project home
    """

    project_home = os.path.join(os.path.expanduser('~'), ".open-MaStR")

    return project_home


def get_data_version_dir():
    """
    Subdirectory of data/ in PROJECTHOME
    
    See :ref:`docs <Project directory>` for configuration of data version.
    
    Returns
    -------
    path-like object
        Absolute path to `PROJECTHOME/data/<data-version>/`
    """
    data_version = get_data_config()["data_version"]
    data_path = os.path.join(get_project_home_dir(), "data", data_version)
    return data_path


def get_filenames():
    """
    Get file names defined in config

    Returns
    -------
    dict
        File names used in open-MaStR
    """
    with open(os.path.join(get_project_home_dir(), "config", "filenames.yml")) as filename_fh:
        filenames = yaml.safe_load(filename_fh)

    return filenames


def get_data_config():
    """
    Get data version and more parameters

    Returns
    -------
    dict
        Configuration parameters
    """
    with open(os.path.join(get_project_home_dir(), "config", "data.yml")) as data_fh:
        data_config = yaml.safe_load(data_fh)

    return data_config


def create_project_home_dir():
    """Create directory structure of PROJECTHOME
    """
    project_home = get_project_home_dir()

    # Create root project home path
    if not os.path.isdir(project_home):
        # Create project home
        log.info('Create {} used for config, parameters and data.'.format(project_home))
        os.mkdir(project_home)

    # Create project home subdirs
    subdirs = [os.path.join(project_home, s) for s in ['config', 'logs', 'data']]
    for subdir in subdirs:
        if not os.path.isdir(subdir):
            os.mkdir(subdir)

    # copy default config files
    config_path = os.path.join(get_project_home_dir(), "config")
    log.info(f'I will create a default set of config files in {config_path}')

    internal_config_dir = os.path.join(pathlib.Path(__file__).parent.absolute(), 'config')
    files = ["data.yml", "logging.yml"]

    for file in files:
        if not file in os.listdir(config_path):
            shutil.copy(os.path.join(internal_config_dir, file),
                        os.path.join(config_path,
                                     os.path.basename(file)))


def create_data_dir():
    """
    Create direct for current data version
    
    The directory that is created for this fata version can be returned by :func:`~.get_data_version_dir`.
    """
    
    os.makedirs(get_data_version_dir(), exist_ok=True)


def get_power_unit_types():
    """
    Returns
    -------
    list
       Technology names 
    """
    return ["wind", "hydro", "solar", "biomass", "combustion", "nuclear", "gsgk", "storage"]


def _filenames_generator():
    """Write default file names .yml to project home dir"""

    filenames_file = os.path.join(get_project_home_dir(), "config", "filenames.yml")

    if not os.path.isfile(filenames_file):

        # How files are prefixed
        prefix = "bnetza_mastr"

        # Additional data available for certain technologies
        type_specific_data = {
            "eeg": ["wind", "hydro", "solar", "biomass", "gsgk", "storage"],
            "kwk": ["combustion", "biomass", "gsgk"],
            "permit": ["wind", "biomass", "combustion", "gsgk", "nuclear", "solar", "hydro"]
        }

        # Template for file names
        filenames_template = {
            "raw": {
                "joined": "{prefix}_{technology}_raw.csv",
                "basic": "{prefix}_{technology}_basic.csv", # power-unit
                "extended": "{prefix}_{technology}_extended.csv", # unit
                "eeg": "{prefix}_{technology}_eeg.csv",
                "kwk": "{prefix}_{technology}_kwk.csv",
                "permit": "{prefix}_{technology}_permit.csv",
                "extended_fail": "{prefix}_{technology}_extended_fail.csv",
                "eeg_fail": "{prefix}_{technology}_eeg_fail.csv",
                "kwk_fail": "{prefix}_{technology}_kwk_fail.csv",
                "permit_fail": "{prefix}_{technology}_permit_fail.csv",
            }
        }

        filenames = {}

        # Define filenames .yml with a dict
        for section, section_filenames in filenames_template.items():
            filenames[section] = {}
            for tech in get_power_unit_types():

                # Files for all technologies
                files = ["joined", "basic", "extended", "extended_fail"]

                # Additional file for some technologies
                for t, techs in type_specific_data.items():
                    if tech in techs:
                        files.append(t)
                        files.append(t + "_fail")

                # Create filename dictionary for one technologies
                    tmp = {
                        k: v.format(prefix=prefix, technology=tech) for k, v in section_filenames.items() if k in files}

                    # Collect file names for all technologies
                    filenames[section].update({tech: tmp})

        # Add file names of cleaned data
        filenames["cleaned"] = {
            tech: f"{prefix}_{tech}_cleaned.csv" for tech in get_power_unit_types()
        }

        # Add file names of processed data
        filenames["postprocessed"] = {
            tech: f"{prefix}_{tech}.csv" for tech in get_power_unit_types()
        }

        # Add filenames for location data
        location_types = ["location_elec_generation", "location_elec_consumption", "location_gas_generation",
            "location_gas_consumption"]
        filenames["raw"].update({loc: f"{loc}_raw.csv" for loc in location_types})

        # Add metadata file
        filenames["metadata"] = "datapackage.json"

        with open(filenames_file, 'w') as outfile:
            yaml.dump(filenames, outfile)
        log.info("File names configuration saved to {}".format(filenames_file))


def setup_project_home():
    """Create open-MaStR project home directory structure

    Create PROJECTHOME returned by :func:`~.get_project_home_dir`.
    In addition, default config files are copied to `PROJECTHOME/config/`.
     """

    # Create directory structure of project home dir
    create_project_home_dir()

    # Save default file names
    _filenames_generator()


def setup_logger():
    """Configure logging in console and log file.
    
    Returns
    -------
    logging.Logger
        Logger with two handlers: console and file.
    """

    # Read logging config
    with open(os.path.join(get_project_home_dir(), "config", "logging.yml")) as filename_fh:
        logging_config = yaml.safe_load(filename_fh)

    # Add logfile location
    logging_config["handlers"]["file"]["filename"] = os.path.join(get_project_home_dir(), "logs", 'open_mastr.log')

    logging.config.dictConfig(logging_config)
    rl = logging.getLogger("open-MaStR")

    return rl


def column_renaming():
    """
    Column renaming for CSV export of raw data

    Helps to export duplicate columns from different data sources.

    Returns
    -------
    dict
        Suffix and column to be suffixed keyed by data type.
    """
    return {
        "basic_data": {
            "columns": ["StatisikFlag", "BestandsanlageMastrNummer"],
            "suffix": "basic"
        },
        "unit_data": {
            "columns": ["EinheitMastrNummer",
                        "EegMastrNummer",
                        "KwkMastrNummer",
                        "GenMastrNummer",
                        "SpeMastrNummer",
                        "EinheitBetriebsstatus",
                        "NichtVorhandenInMigriertenEinheiten",
                        "Bruttoleistung"],
            "suffix": "extended"
        },
        "eeg_data": {
            "columns": ["EegMastrNummer", "DatumLetzteAktualisierung", "Meldedatum"],
            "suffix": "eeg"
        },
        "kwk_data": {
            "columns": ["KwkMastrNummer",
                        "Meldedatum",
                        "Inbetriebnahmedatum",
                        "DatumLetzteAktualisierung",
                        "AnlageBetriebsstatus",
                        "VerknuepfteEinheiten",
                        "AusschreibungZuschlag"],
            "suffix": "kwk"
        },
        "permit_data": {
            "columns": ["GenMastrNummer", "DatumLetzteAktualisierung", "Meldedatum"],
            "suffix": "permit"
        },
    }



from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.schema import MetaData
from sqlalchemy import (
    Column,
    Integer,
    String,
    Float,
    Sequence,
    DateTime,
    Boolean,
    func,
    Date,
    JSON,
)
import os

DB_ENGINE = os.environ.get("DB_ENGINE", "sqlite")


mirror_schema = "mastr_mirrored" if DB_ENGINE == "docker" else None
meta = MetaData(schema=mirror_schema)
Base = declarative_base(metadata=meta)


class ParentAllTables(object):

    DatenQuelle = Column(String)
    DatumDownload = Column(Date)


class BasicUnit(Base):
    __tablename__ = "basic_units"

    EinheitMastrNummer = Column(String, primary_key=True)
    DatumLetzteAktualisierung = Column(DateTime(timezone=True))
    Name = Column(String)
    Einheitart = Column(String)
    Einheittyp = Column(String)
    Standort = Column(String)
    Bruttoleistung = Column(Float)
    Erzeugungsleistung = Column(Float)
    EinheitBetriebsstatus = Column(String)
    Anlagenbetreiber = Column(String)
    EegMastrNummer = Column(String)
    KwkMastrNummer = Column(String)
    SpeMastrNummer = Column(String)
    GenMastrNummer = Column(String)
    BestandsanlageMastrNummer = Column(String)
    NichtVorhandenInMigriertenEinheiten = Column(String)
    StatisikFlag = Column(String)


class AdditionalDataRequested(Base):
    __tablename__ = "additional_data_requested"

    id = Column(
        Integer,
        Sequence("additional_data_requested_id_seq", schema=mirror_schema),
        primary_key=True,
    )
    EinheitMastrNummer = Column(String)
    additional_data_id = Column(String)
    technology = Column(String)
    data_type = Column(String)
    request_date = Column(DateTime(timezone=True), default=func.now())


class MissedAdditionalData(Base):
    __tablename__ = "missed_additional_data"

    id = Column(
        Integer,
        Sequence("additional_data_missed_id_seq", schema=mirror_schema),
        primary_key=True,
    )
    additional_data_id = Column(String)
    reason = Column(String)
    download_date = Column(DateTime(timezone=True), default=func.now())


class Extended(object):

    NetzbetreiberMastrNummer = Column(String)
    Registrierungsdatum = Column(Date)
    EinheitMastrNummer = Column(String, primary_key=True)
    DatumLetzteAktualisierung = Column(DateTime(timezone=True))
    LokationMastrNummer = Column(String)
    NetzbetreiberpruefungStatus = Column(String)
    NetzbetreiberpruefungDatum = Column(Date)
    AnlagenbetreiberMastrNummer = Column(String)
    Land = Column(String)
    Bundesland = Column(String)
    Landkreis = Column(String)
    Gemeinde = Column(String)
    Gemeindeschluessel = Column(String)
    Postleitzahl = Column(String)
    Gemarkung = Column(String)
    FlurFlurstuecknummern = Column(String)
    Strasse = Column(String)
    StrasseNichtGefunden = Column(Boolean)
    Hausnummer = Column(String)
    HausnummerNichtGefunden = Column(Boolean)
    Adresszusatz = Column(String)
    Ort = Column(String)
    Laengengrad = Column(String) if DB_ENGINE == "docker" else Column(Float)
    Breitengrad = Column(String) if DB_ENGINE == "docker" else Column(Float)
    UtmZonenwert = Column(String)
    UtmEast = Column(String) if DB_ENGINE == "docker" else Column(Float)
    UtmNorth = Column(String) if DB_ENGINE == "docker" else Column(Float)
    GaussKruegerHoch = Column(String) if DB_ENGINE == "docker" else Column(Float)
    GaussKruegerRechts = Column(String) if DB_ENGINE == "docker" else Column(Float)
    Meldedatum = Column(Date)
    GeplantesInbetriebnahmedatum = Column(Date)
    Inbetriebnahmedatum = Column(Date)
    DatumEndgueltigeStilllegung = Column(Date)
    DatumBeginnVoruebergehendeStilllegung = Column(Date)
    DatumBeendigungVorlaeufigenStilllegung = Column(Date)
    DatumWiederaufnahmeBetrieb = Column(Date)
    EinheitSystemstatus = Column(String)
    EinheitBetriebsstatus = Column(String)
    BestandsanlageMastrNummer = Column(String)
    NichtVorhandenInMigriertenEinheiten = Column(Boolean)
    AltAnlagenbetreiberMastrNummer = Column(String)
    DatumDesBetreiberwechsels = Column(Date)
    DatumRegistrierungDesBetreiberwechsels = Column(Date)
    StatisikFlag = Column(String)
    NameStromerzeugungseinheit = Column(String)
    Weic = Column(String)
    WeicDisplayName = Column(String)
    Kraftwerksnummer = Column(String)
    Energietraeger = Column(String)
    Bruttoleistung = Column(Float)
    Nettonennleistung = Column(Float)
    AnschlussAnHoechstOderHochSpannung = Column(Boolean)
    Schwarzstartfaehigkeit = Column(Boolean)
    Inselbetriebsfaehigkeit = Column(Boolean)
    Einsatzverantwortlicher = Column(String)
    FernsteuerbarkeitNb = Column(Boolean)
    FernsteuerbarkeitDv = Column(Boolean)
    FernsteuerbarkeitDr = Column(Boolean)
    Einspeisungsart = Column(String)
    PraequalifiziertFuerRegelenergie = Column(Boolean)
    GenMastrNummer = Column(String)
    # from bulk download
    Hausnummer_nv = Column(Boolean)
    Weic_nv = Column(Boolean)
    Kraftwerksnummer_nv = Column(Boolean)
    NichtVorhandenInMigriertenEinheiten = Column(Boolean)
    Einsatzverantwortlicher = Column(String)


class WindExtended(Extended, ParentAllTables, Base):
    __tablename__ = "wind_extended"

    # wind specific attributes
    NameWindpark = Column(String)
    Lage = Column(String)
    Seelage = Column(String)
    ClusterOstsee = Column(String)
    ClusterNordsee = Column(String)
    Hersteller = Column(String)
    HerstellerId = Column(String)
    Technologie = Column(String)
    Typenbezeichnung = Column(String)
    Nabenhoehe = Column(Float)
    Rotordurchmesser = Column(Float)
    Rotorblattenteisungssystem = Column(Boolean)
    AuflageAbschaltungLeistungsbegrenzung = Column(Boolean)
    AuflagenAbschaltungSchallimmissionsschutzNachts = Column(Boolean)
    AuflagenAbschaltungSchallimmissionsschutzTagsueber = Column(Boolean)
    AuflagenAbschaltungSchattenwurf = Column(Boolean)
    AuflagenAbschaltungTierschutz = Column(Boolean)
    AuflagenAbschaltungEiswurf = Column(Boolean)
    AuflagenAbschaltungSonstige = Column(Boolean)
    Wassertiefe = Column(Float)
    Kuestenentfernung = Column(Float)
    EegMastrNummer = Column(String)


class SolarExtended(Extended, ParentAllTables, Base):
    __tablename__ = "solar_extended"

    zugeordneteWirkleistungWechselrichter = Column(Float)
    GemeinsamerWechselrichterMitSpeicher = Column(String)
    AnzahlModule = Column(Integer)
    Lage = Column(String)
    Leistungsbegrenzung = Column(String)
    EinheitlicheAusrichtungUndNeigungswinkel = Column(Boolean)
    Hauptausrichtung = Column(String)
    HauptausrichtungNeigungswinkel = Column(String)
    Nebenausrichtung = Column(String)
    NebenausrichtungNeigungswinkel = Column(String)
    InAnspruchGenommeneFlaeche = Column(Float)
    ArtDerFlaeche = Column(String)
    InAnspruchGenommeneAckerflaeche = Column(Float)
    Nutzungsbereich = Column(String)
    EegMastrNummer = Column(String)
    ArtDerFlaecheIds = Column(String)


class BiomassExtended(Extended, ParentAllTables, Base):
    __tablename__ = "biomass_extended"

    Hauptbrennstoff = Column(String)
    Biomasseart = Column(String)
    Technologie = Column(String)
    EegMastrNummer = Column(String)
    KwkMastrNummer = Column(String)


class CombustionExtended(Extended, ParentAllTables, Base):
    __tablename__ = "combustion_extended"

    NameKraftwerk = Column(String)
    NameKraftwerksblock = Column(String)
    DatumBaubeginn = Column(Date)
    AnzeigeEinerStilllegung = Column(Boolean)
    ArtDerStilllegung = Column(String)
    DatumBeginnVorlaeufigenOderEndgueltigenStilllegung = Column(Date)
    SteigerungNettonennleistungKombibetrieb = Column(Float)
    AnlageIstImKombibetrieb = Column(Boolean)
    MastrNummernKombibetrieb = Column(String)
    NetzreserveAbDatum = Column(Date)
    SicherheitsbereitschaftAbDatum = Column(Date)
    Hauptbrennstoff = Column(String)
    WeitererHauptbrennstoff = Column(String)
    WeitereBrennstoffe = Column(String)
    VerknuepfteErzeugungseinheiten = Column(String)
    BestandteilGrenzkraftwerk = Column(Boolean)
    NettonennleistungDeutschland = Column(Float)
    AnteiligNutzungsberechtigte = Column(String)
    Notstromaggregat = Column(Boolean)
    Einsatzort = Column(String)
    KwkMastrNummer = Column(String)
    Technologie = Column(String)


class GsgkExtended(Extended, ParentAllTables, Base):
    __tablename__ = "gsgk_extended"

    Technologie = Column(String)
    KwkMastrNummer = Column(String)
    EegMastrNummer = Column(String)


class HydroExtended(Extended, ParentAllTables, Base):
    __tablename__ = "hydro_extended"

    NameKraftwerk = Column(String)
    ArtDerWasserkraftanlage = Column(String)
    AnzeigeEinerStilllegung = Column(Boolean)
    ArtDerStilllegung = Column(String)
    DatumBeginnVorlaeufigenOderEndgueltigenStilllegung = Column(Date)
    MinderungStromerzeugung = Column(Boolean)
    BestandteilGrenzkraftwerk = Column(Boolean)
    NettonennleistungDeutschland = Column(Float)
    ArtDesZuflusses = Column(String)
    EegMastrNummer = Column(String)


class NuclearExtended(Extended, ParentAllTables, Base):
    __tablename__ = "nuclear_extended"

    NameKraftwerk = Column(String)
    NameKraftwerksblock = Column(String)
    Technologie = Column(String)


class StorageExtended(Extended, ParentAllTables, Base):
    __tablename__ = "storage_extended"

    Einsatzort = Column(String)
    AcDcKoppelung = Column(String)
    Batterietechnologie = Column(String)
    PumpbetriebLeistungsaufnahme = Column(Float)
    PumpbetriebKontinuierlichRegelbar = Column(Boolean)
    Pumpspeichertechnologie = Column(String)
    Notstromaggregat = Column(Boolean)
    BestandteilGrenzkraftwerk = Column(Boolean)
    NettonennleistungDeutschland = Column(Float)
    ZugeordnenteWirkleistungWechselrichter = Column(Float)
    NutzbareSpeicherkapazitaet = Column(Float)
    SpeMastrNummer = Column(String)
    EegMastrNummer = Column(String)
    EegAnlagentyp = Column(String)
    Technologie = Column(String)


class Eeg(object):

    Registrierungsdatum = Column(Date)
    EegMastrNummer = Column(String, primary_key=True)
    Meldedatum = Column(Date)
    DatumLetzteAktualisierung = Column(DateTime(timezone=True))
    EegInbetriebnahmedatum = Column(Date)
    VerknuepfteEinheit = Column(String)


class WindEeg(Eeg, ParentAllTables, Base):
    __tablename__ = "wind_eeg"

    AnlagenkennzifferAnlagenregister = Column(String)
    AnlagenschluesselEeg = Column(String)
    PrototypAnlage = Column(Boolean)
    PilotAnlage = Column(Boolean)
    InstallierteLeistung = Column(Float)
    VerhaeltnisErtragsschaetzungReferenzertrag = Column(Float)
    VerhaeltnisReferenzertragErtrag5Jahre = Column(Float)
    VerhaeltnisReferenzertragErtrag10Jahre = Column(Float)
    VerhaeltnisReferenzertragErtrag15Jahre = Column(Float)
    AusschreibungZuschlag = Column(Boolean)
    Zuschlagsnummer = Column(String)
    AnlageBetriebsstatus = Column(String)
    AnlagenkennzifferAnlagenregister_nv = Column(Boolean)
    VerhaeltnisErtragsschaetzungReferenzertrag_nv = Column(Boolean)
    VerhaeltnisReferenzertragErtrag5Jahre_nv = Column(Boolean)
    VerhaeltnisReferenzertragErtrag10Jahre_nv = Column(Boolean)
    VerhaeltnisReferenzertragErtrag15Jahre_nv = Column(Boolean)


class SolarEeg(Eeg, ParentAllTables, Base):
    __tablename__ = "solar_eeg"

    InanspruchnahmeZahlungNachEeg = Column(Boolean)
    AnlagenschluesselEeg = Column(String)
    AnlagenkennzifferAnlagenregister = Column(String)
    InstallierteLeistung = Column(Float)
    RegistrierungsnummerPvMeldeportal = Column(String)
    MieterstromRegistrierungsdatum = Column(Date)
    MieterstromZugeordnet = Column(Boolean)
    MieterstromMeldedatum = Column(Date)
    MieterstromErsteZuordnungZuschlag = Column(Date)
    AusschreibungZuschlag = Column(Boolean)
    ZugeordneteGebotsmenge = Column(Float)
    Zuschlagsnummer = Column(String)
    AnlageBetriebsstatus = Column(String)
    AnlagenkennzifferAnlagenregister_nv = Column(Boolean)
    RegistrierungsnummerPvMeldeportal_nv = Column(Boolean)


class BiomassEeg(Eeg, ParentAllTables, Base):
    __tablename__ = "biomass_eeg"

    AnlagenschluesselEeg = Column(String)
    AnlagenkennzifferAnlagenregister = Column(String)
    InstallierteLeistung = Column(Float)
    AusschliesslicheVerwendungBiomasse = Column(Boolean)
    AusschreibungZuschlag = Column(Boolean)
    Zuschlagsnummer = Column(String)
    BiogasInanspruchnahmeFlexiPraemie = Column(Boolean)
    BiogasDatumInanspruchnahmeFlexiPraemie = Column(Date)
    BiogasLeistungserhoehung = Column(Boolean)
    BiogasDatumLeistungserhoehung = Column(Date)
    BiogasUmfangLeistungserhoehung = Column(Float)
    BiogasGaserzeugungskapazitaet = Column(Float)
    BiogasHoechstbemessungsleistung = Column(Float)
    BiomethanErstmaligerEinsatz = Column(Date)
    AnlageBetriebsstatus = Column(String)
    AnlagenkennzifferAnlagenregister_nv = Column(Boolean)
    BiogasGaserzeugungskapazitaet_nv = Column(Boolean)
    BiomethanErstmaligerEinsatz_nv = Column(Boolean)


class GsgkEeg(Eeg, ParentAllTables, Base):
    __tablename__ = "gsgk_eeg"

    AnlagenschluesselEeg = Column(String)
    AnlagenkennzifferAnlagenregister = Column(String)
    InstallierteLeistung = Column(Float)
    AnlageBetriebsstatus = Column(String)
    AnlagenkennzifferAnlagenregister_nv = Column(Boolean)


class HydroEeg(Eeg, ParentAllTables, Base):
    __tablename__ = "hydro_eeg"

    AnlagenschluesselEeg = Column(String)
    AnlagenkennzifferAnlagenregister = Column(String)
    InstallierteLeistung = Column(Float)
    AnlageBetriebsstatus = Column(String)
    Ertuechtigung = Column(JSON)
    AnlagenkennzifferAnlagenregister_nv = Column(Boolean)
    ErtuechtigungIds = Column(String)


class StorageEeg(Eeg, ParentAllTables, Base):
    __tablename__ = "storage_eeg"


class Kwk(ParentAllTables, Base):
    __tablename__ = "kwk"

    Registrierungsdatum = Column(Date)
    KwkMastrNummer = Column(String, primary_key=True)
    AusschreibungZuschlag = Column(Boolean)
    Zuschlagnummer = Column(String)
    DatumLetzteAktualisierung = Column(DateTime(timezone=True))
    Inbetriebnahmedatum = Column(Date)
    Meldedatum = Column(Date)
    ThermischeNutzleistung = Column(Float)
    ElektrischeKwkLeistung = Column(Float)
    VerknuepfteEinheiten = Column(String)
    AnlageBetriebsstatus = Column(String)


class Permit(ParentAllTables, Base):
    __tablename__ = "permit"

    Registrierungsdatum = Column(Date)
    GenMastrNummer = Column(String, primary_key=True)
    DatumLetzteAktualisierung = Column(DateTime(timezone=True))
    Art = Column(String)
    Datum = Column(Date)
    Behoerde = Column(String)
    Aktenzeichen = Column(String)
    Frist = Column(Date)
    WasserrechtsNummer = Column(String)
    WasserrechtAblaufdatum = Column(Date)
    Meldedatum = Column(Date)
    VerknuepfteEinheiten = Column(String)
    Frist_nv = Column(Boolean)
    WasserrechtAblaufdatum_nv = Column(Boolean)


class LocationBasic(ParentAllTables, Base):
    __tablename__ = "locations_basic"

    LokationMastrNummer = Column(String, primary_key=True)
    NameDerTechnischenLokation = Column(String)
    Lokationtyp = Column(String)
    AnzahlNetzanschlusspunkte = Column(Integer)


class LocationExtended(ParentAllTables, Base):
    __tablename__ = "locations_extended"

    MastrNummer = Column(String, primary_key=True)
    DatumLetzteAktualisierung = Column(DateTime(timezone=True))
    NameDerTechnischenLokation = Column(String)
    VerknuepfteEinheiten = Column(JSONB) if DB_ENGINE == "docker" else Column(JSON)
    Netzanschlusspunkte = Column(JSONB) if DB_ENGINE == "docker" else Column(JSON)
    Lokationtyp = Column(String)


class AdditionalLocationsRequested(ParentAllTables, Base):
    __tablename__ = "additional_locations_requested"

    id = Column(
        Integer,
        Sequence("additional_locations_requested_id_seq", schema=mirror_schema),
        primary_key=True,
    )
    LokationMastrNummer = Column(String)
    location_type = Column(String)
    request_date = Column(DateTime(timezone=True), default=func.now())


class MissedExtendedLocation(ParentAllTables, Base):
    __tablename__ = "missed_extended_location_data"

    id = Column(
        Integer,
        Sequence("additional_location_data_missed_id_seq", schema=mirror_schema),
        primary_key=True,
    )
    LokationMastrNummer = Column(String)
    reason = Column(String)
    download_date = Column(DateTime(timezone=True), default=func.now())


class GasStorage(ParentAllTables, Base):
    __tablename__ = "gas_storage"

    MaStRNummer = Column(String, primary_key=True)
    DatumLetzteAktualisierung = Column(DateTime(timezone=True))
    Speichername = Column(String)
    Registrierungsdatum = Column(Date)
    AnlageBetriebsstatus = Column(String)
    VerknuepfteEinheit = Column(String)


class GasStorageExtended(ParentAllTables, Base):
    __tablename__ = "gas_storage_extended"
    EinheitMastrNummer = Column(String, primary_key=True)
    DatumLetzteAktualisierung = Column(DateTime(timezone=True))
    LokationMaStRNummer = Column(String)
    NetzbetreiberpruefungStatus = Column(Boolean)
    NetzbetreiberpruefungDatum = Column(Date)
    AnlagenbetreiberMastrNummer = Column(String)
    Land = Column(String)
    Bundesland = Column(String)
    Landkreis = Column(String)
    Gemeinde = Column(String)
    Gemeindeschluessel = Column(String)
    Postleitzahl = Column(Integer)
    Ort = Column(String)
    Strasse = Column(String)
    StrasseNichtGefunden = Column(Integer)
    Hausnummer = Column(String)
    Hausnummer_nv = Column(Integer)
    HausnummerNichtGefunden = Column(Integer)
    Laengengrad = Column(Float)
    Breitengrad = Column(Float)
    Registrierungsdatum = Column(String)
    Inbetriebnahmedatum = Column(String)
    EinheitSystemstatus = Column(String)
    EinheitBetriebsstatus = Column(String)
    NichtVorhandenInMigriertenEinheiten = Column(Integer)
    NameGasspeicher = Column(String)
    Speicherart = Column(String)
    MaximalNutzbaresArbeitsgasvolumen = Column(Float)
    MaximaleEinspeicherleistung = Column(Float)
    MaximaleAusspeicherleistung = Column(Float)
    DurchschnittlicherBrennwert = Column(Float)
    Weic = Column(String)
    Weic_Na = Column(Integer)
    SpeicherMaStRNummer = Column(String)
    Gemarkung = Column(String)
    FlurFlurstuecknummern = Column(String)
    Adresszusatz = Column(String)
    DatumBeginnVoruebergehendeStilllegung = Column(Date)


class StorageUnits(ParentAllTables, Base):
    __tablename__ = "storage_units"
    MaStRNummer = Column(String, primary_key=True)
    Registrierungsdatum = Column(Date)
    DatumLetzteAktualisierung = Column(DateTime(timezone=True))
    NutzbareSpeicherkapazitaet = Column(Float)
    VerknuepfteEinheit = Column(String)
    AnlageBetriebsstatus = Column(String)


class BalancingArea(ParentAllTables, Base):
    __tablename__ = "balancing_area"

    Id = Column(Integer, primary_key=True)
    Yeic = Column(String)
    RegelzoneNetzanschlusspunkt = Column(String)
    BilanzierungsgebietNetzanschlusspunkt = Column(String)


class GasProducer(ParentAllTables, Base):
    __tablename__ = "gas_producer"

    EinheitMaStRNummer = Column(String, primary_key=True)
    DatumLetzteAktualisierung = Column(DateTime(timezone=True))
    LokationMaStRNummer = Column(String)
    NetzbetreiberpruefungStatus = Column(Boolean)
    NetzbetreiberpruefungDatum = Column(Date)
    AnlagenbetreiberMastrNummer = Column(String)
    Land = Column(String)
    Bundesland = Column(String)
    Landkreis = Column(String)
    Gemeinde = Column(String)
    Gemeindeschluessel = Column(String)
    Postleitzahl = Column(Integer)
    Ort = Column(String)
    Registrierungsdatum = Column(Date)
    Inbetriebnahmedatum = Column(Date)
    EinheitSystemstatus = Column(String)
    EinheitBetriebsstatus = Column(String)
    NichtVorhandenInMigriertenEinheiten = Column(Integer)
    NameGaserzeugungseinheit = Column(String)
    SpeicherMaStRNummer = Column(String)
    Strasse = Column(String)
    StrasseNichtGefunden = Column(Integer)
    Hausnummer = Column(String)
    Hausnummer_nv = Column(Integer)
    HausnummerNichtGefunden = Column(Integer)
    Adresszusatz = Column(String)
    Laengengrad = Column(Float)
    Breitengrad = Column(Float)
    Technologie = Column(String)
    Erzeugungsleistung = Column(Float)
    DatumDesBetreiberwechsels = Column(Date)
    DatumRegistrierungDesBetreiberwechsels = Column(Date)
    Gemarkung = Column(String)
    FlurFlurstuecknummern = Column(String)
    GeplantesInbetriebnahmedatum = Column(Date)
    DatumBeginnVoruebergehendeStilllegung = Column(Date)


class GasConsumer(ParentAllTables, Base):
    __tablename__ = "gas_consumer"

    EinheitMaStRNummer = Column(String, primary_key=True)
    DatumLetzteAktualisierung = Column(DateTime(timezone=True))
    LokationMaStRNummer = Column(String)
    NetzbetreiberpruefungStatus = Column(Boolean)
    NetzbetreiberpruefungDatum = Column(Date)
    AnlagenbetreiberMastrNummer = Column(String)
    Land = Column(String)
    Bundesland = Column(String)
    Landkreis = Column(String)
    Gemeinde = Column(String)
    Gemeindeschluessel = Column(String)
    Postleitzahl = Column(Integer)
    Ort = Column(String)
    Strasse = Column(String)
    StrasseNichtGefunden = Column(Integer)
    Hausnummer = Column(String)
    Hausnummer_nv = Column(Integer)
    HausnummerNichtGefunden = Column(Integer)
    Laengengrad = Column(Float)
    Breitengrad = Column(Float)
    Registrierungsdatum = Column(String)
    Inbetriebnahmedatum = Column(String)
    EinheitSystemstatus = Column(String)
    EinheitBetriebsstatus = Column(String)
    NichtVorhandenInMigriertenEinheiten = Column(Integer)
    NameGasverbrauchsseinheit = Column(String)
    EinheitDientDerStromerzeugung = Column(String)
    MaximaleGasbezugsleistung = Column(Float)
    VerknuepfteEinheit = Column(String)
    GeplantesInbetriebnahmedatum = Column(Date)
    Adresszusatz = Column(String)
    Gemarkung = Column(String)
    FlurFlurstuecknummern = Column(String)
    DatumDesBetreiberwechsels = Column(Date)
    DatumRegistrierungDesBetreiberwechsels = Column(Date)
    DatumEndgueltigeStilllegung = Column(Date)
    DatumBeginnVoruebergehendeStilllegung = Column(Date)


class ElectricityConsumer(ParentAllTables, Base):
    __tablename__ = "electricity_consumer"

    EinheitMaStRNummer = Column(String, primary_key=True)
    DatumLetzteAktualisierung = Column(DateTime(timezone=True))
    LokationMaStRNummer = Column(String)
    NetzbetreiberpruefungStatus = Column(Boolean)
    NetzbetreiberpruefungDatum = Column(Date)
    AnlagenbetreiberMastrNummer = Column(String)
    Land = Column(String)
    Bundesland = Column(String)
    Landkreis = Column(String)
    Gemeinde = Column(String)
    Gemeindeschluessel = Column(String)
    Postleitzahl = Column(Integer)
    Ort = Column(String)
    Strasse = Column(String)
    StrasseNichtGefunden = Column(Integer)
    Hausnummer = Column(String)
    Hausnummer_nv = Column(Integer)
    HausnummerNichtGefunden = Column(Integer)
    Adresszusatz = Column(String)
    Gemarkung = Column(String)
    FlurFlurstuecknummern = Column(String)
    Laengengrad = Column(Float)
    Breitengrad = Column(Float)
    Registrierungsdatum = Column(String)
    Inbetriebnahmedatum = Column(String)
    EinheitSystemstatus = Column(String)
    EinheitBetriebsstatus = Column(String)
    NichtVorhandenInMigriertenEinheiten = Column(Integer)
    Einsatzverantwortlicher = Column(String)
    NameStromverbrauchseinheit = Column(String)
    AnzahlStromverbrauchseinheitenGroesser50Mw = Column(Integer)
    PraequalifiziertGemaessAblav = Column(Boolean)
    AnteilBeinflussbareLast = Column(Float)
    ArtAbschaltbareLast = Column(String)
    DatumDesBetreiberwechsels = Column(Date)
    DatumRegistrierungDesBetreiberwechsels = Column(Date)
    DatumEndgueltigeStilllegung = Column(Date)
    GeplantesInbetriebnahmedatum = Column(Date)


class MarketRoles(ParentAllTables, Base):
    __tablename__ = "market_roles"

    MastrNummer = Column(String, primary_key=True)
    MarktakteurMastrNummer = Column(String)
    Marktrolle = Column(String)
    Marktpartneridentifikationsnummer_nv = Column(Boolean)
    BundesnetzagenturBetriebsnummer = Column(String)
    BundesnetzagenturBetriebsnummer_nv = Column(Boolean)
    Marktpartneridentifikationsnummer = Column(String)
    KontaktdatenMarktrolle = Column(String)
    DatumLetzteAktualisierung = Column(DateTime(timezone=True))


class MarketActors(ParentAllTables, Base):
    __tablename__ = "market_actors"

    MastrNummer = Column(String, primary_key=True)
    Personenart = Column(String)
    Marktfunktion = Column(String)
    RegistergerichtAusland = Column(String)
    Registernummer = Column(String)
    DatumLetzeAktualisierung = Column(DateTime(timezone=True))
    Firmenname = Column(String)
    Rechtsform = Column(String)
    Land = Column(String)
    Strasse = Column(String)
    Hausnummer = Column(String)
    Hausnummer_nv = Column(Boolean)
    Postleitzahl = Column(String)
    Ort = Column(String)
    Bundesland = Column(String)
    Nuts2 = Column(String)
    Email = Column(String)
    Telefon = Column(String)
    Fax_nv = Column(Boolean)
    Webseite_nv = Column(Boolean)
    Taetigkeitsbeginn = Column(Date)
    AcerCode_nv = Column(Boolean)
    Umsatzsteueridentifikationsnummer_nv = Column(Boolean)
    BundesnetzagenturBetriebsnummer = Column(String)
    BundesnetzagenturBetriebsnummer_nv = Column(Boolean)
    HausnummerAnZustelladresse_nv = Column(Boolean)
    Kmu = Column(Integer)
    RegistrierungsdatumMarktakteur = Column(DateTime(timezone=True))
    Fax = Column(String)
    HauptwirtdschaftszweigAbteilung = Column(String)
    HauptwirtdschaftszweigGruppe = Column(String)
    HauptwirtdschaftszweigAbschnitt = Column(String)
    Webseite = Column(String)
    Umsatzsteueridentifikationsnummer = Column(String)
    Registergericht = Column(String)
    Adresszusatz = Column(String)
    LandAnZustelladresse = Column(String)
    PostleitzahlAnZustelladresse = Column(String)
    OrtAnZustelladresse = Column(String)
    StrasseAnZustelladresse = Column(String)
    HausnummerAnZustelladresse = Column(String)
    RegisternummerAusland = Column(String)
    SonstigeRechtsform = Column(String)
    AcerCode = Column(String)
    AdresszusatzAnZustelladresse = Column(String)
    Taetigkeitsende = Column(Date)
    Region = Column(String)
    Taetigkeitsende_nv = Column(Boolean)
    Marktrollen = Column(String)
    Gasgrosshaendler = Column(Boolean)
    BelieferungVonLetztverbrauchernGas = Column(Boolean)
    BelieferungHaushaltskundenGas = Column(Boolean)
    Netz = Column(String)
    Direktvermarktungsunternehmen = Column(Boolean)
    BelieferungVonLetztverbrauchernStrom = Column(Boolean)
    BelieferungHaushaltskundenStrom = Column(Boolean)
    Stromgrosshaendler = Column(Boolean)
    MarktakteurVorname = Column(String)
    MarktakteurNachname = Column(String)


class Grids(ParentAllTables, Base):
    __tablename__ = "grids"

    MastrNummer = Column(String, primary_key=True)
    DatumLetzteAktualisierung = Column(DateTime(timezone=True))
    Sparte = Column(String)
    KundenAngeschlossen = Column(String)
    GeschlossenesVerteilnetz = Column(String)
    Bezeichnung = Column(String)
    Marktgebiet = Column(String)


class GridConnections(ParentAllTables, Base):
    __tablename__ = "grid_connections"

    NetzanschlusspunktMastrNummer = Column(String, primary_key=True)
    NetzanschlusspunktBezeichnung = Column(String)
    LetzteAenderung = Column(DateTime(timezone=True))
    LokationMaStRNummer = Column(String)
    Lokationtyp = Column(String)
    MaximaleEinspeiseleistung = Column(Float)
    Gasqualitaet = Column(String)
    NetzMaStRNummer = Column(String)
    NochInPlanung = Column(Boolean)
    NameDerTechnischenLokation = Column(String)
    MaximaleAusspeiseleistung = Column(Float)
    Messlokation = Column(String)
    Spannungsebene = Column(String)
    BilanzierungsgebietNetzanschlusspunktId = Column(Integer)
    Nettoengpassleistung = Column(Float)
    Netzanschlusskapazitaet = Column(Float)


tablename_mapping = {
    "anlageneegbiomasse": {
        "__name__": BiomassEeg.__tablename__,
        "__class__": BiomassEeg,
        "replace_column_names": {
            "VerknuepfteEinheitenMaStRNummern": "VerknuepfteEinheit"
        },
    },
    "einheitenbiomasse": {
        "__name__": BiomassExtended.__tablename__,
        "__class__": BiomassExtended,
        "replace_column_names": None,
    },
    "anlageneeggeosolarthermiegrubenklaerschlammdruckentspannung": {
        "__name__": GsgkEeg.__tablename__,
        "__class__": GsgkEeg,
        "replace_column_names": {
            "VerknuepfteEinheitenMaStRNummern": "VerknuepfteEinheit"
        },
    },
    "einheitengeosolarthermiegrubenklaerschlammdruckentspannung": {
        "__name__": GsgkExtended.__tablename__,
        "__class__": GsgkExtended,
        "replace_column_names": None,
    },
    "anlageneegsolar": {
        "__name__": SolarEeg.__tablename__,
        "__class__": SolarEeg,
        "replace_column_names": {
            "VerknuepfteEinheitenMaStRNummern": "VerknuepfteEinheit"
        },
    },
    "einheitensolar": {
        "__name__": SolarExtended.__tablename__,
        "__class__": SolarExtended,
        "replace_column_names": {
            "VerknuepfteEinheitenMaStRNummern": "VerknuepfteEinheit"
        },
    },
    "anlageneegspeicher": {
        "__name__": StorageEeg.__tablename__,
        "__class__": StorageEeg,
        "replace_column_names": {
            "VerknuepfteEinheitenMaStRNummern": "VerknuepfteEinheit"
        },
    },
    "anlageneegwasser": {
        "__name__": HydroEeg.__tablename__,
        "__class__": HydroEeg,
        "replace_column_names": {
            "VerknuepfteEinheitenMaStRNummern": "VerknuepfteEinheit"
        },
    },
    "einheitenwasser": {
        "__name__": HydroExtended.__tablename__,
        "__class__": HydroExtended,
        "replace_column_names": None,
    },
    "anlageneegwind": {
        "__name__": WindEeg.__tablename__,
        "__class__": WindEeg,
        "replace_column_names": {
            "VerknuepfteEinheitenMaStRNummern": "VerknuepfteEinheit"
        },
    },
    "einheitenwind": {
        "__name__": WindExtended.__tablename__,
        "__class__": WindExtended,
        "replace_column_names": None,
    },
    "anlagengasspeicher": {
        "__name__": GasStorage.__tablename__,
        "__class__": GasStorage,
        "replace_column_names": {
            "VerknuepfteEinheitenMaStRNummern": "VerknuepfteEinheit"
        },
    },
    "einheitengasspeicher": {
        "__name__": GasStorageExtended.__tablename__,
        "__class__": GasStorageExtended,
        "replace_column_names": None,
    },
    "anlagenkwk": {
        "__name__": Kwk.__tablename__,
        "__class__": Kwk,
        "replace_column_names": {
            "VerknuepfteEinheitenMaStRNummern": "VerknuepfteEinheiten"
        },
    },
    "anlagenstromspeicher": {
        "__name__": StorageUnits.__tablename__,
        "__class__": StorageUnits,
        "replace_column_names": {
            "VerknuepfteEinheitenMaStRNummern": "VerknuepfteEinheit"
        },
    },
    "bilanzierungsgebiete": {
        "__name__": BalancingArea.__tablename__,
        "__class__": BalancingArea,
        "replace_column_names": None,
    },
    "einheitengaserzeuger": {
        "__name__": GasProducer.__tablename__,
        "__class__": GasProducer,
        "replace_column_names": None,
    },
    "einheitengasverbraucher": {
        "__name__": GasConsumer.__tablename__,
        "__class__": GasConsumer,
        "replace_column_names": {
            "VerknuepfteEinheitenMaStRNummern": "VerknuepfteEinheit"
        },
    },
    "einheitengenehmigung": {
        "__name__": Permit.__tablename__,
        "__class__": Permit,
        "replace_column_names": {
            "VerknuepfteEinheitenMaStRNummern": "VerknuepfteEinheiten"
        },
    },
    "einheitenkernkraft": {
        "__name__": NuclearExtended.__tablename__,
        "__class__": NuclearExtended,
        "replace_column_names": None,
    },
    "einheitenstromverbraucher": {
        "__name__": ElectricityConsumer.__tablename__,
        "__class__": ElectricityConsumer,
        "replace_column_names": None,
    },
    "einheitenstromspeicher": {
        "__name__": StorageExtended.__tablename__,
        "__class__": StorageExtended,
        "replace_column_names": None,
    },
    "einheitenverbrennung": {
        "__name__": CombustionExtended.__tablename__,
        "__class__": CombustionExtended,
        "replace_column_names": None,
    },
    "marktrollen": {
        "__name__": MarketRoles.__tablename__,
        "__class__": MarketRoles,
        "replace_column_names": None,
    },
    "marktakteure": {
        "__name__": MarketActors.__tablename__,
        "__class__": MarketActors,
        "replace_column_names": None,
    },
    "netze": {
        "__name__": Grids.__tablename__,
        "__class__": Grids,
        "replace_column_names": None,
    },
    "netzanschlusspunkte": {
        "__name__": GridConnections.__tablename__,
        "__class__": GridConnections,
        "replace_column_names": None,
    },
    "katalogkategorien": {
        "__name__": "katalogkategorien",
        "__class__": None,
        "replace_column_names": None,
    },
    "katalogwerte": {
        "__name__": "katalogwerte",
        "__class__": None,
        "replace_column_names": None,
    },
    "lokationen": {
        "__name__": LocationExtended.__tablename__,
        "__class__": LocationExtended,
        "replace_column_names": {
            "VerknuepfteEinheitenMaStRNummern": "VerknuepfteEinheiten",
            "NetzanschlusspunkteMaStRNummern": "Netzanschlusspunkte",
        },
    },
    "einheitentypen": {
        "__name__": "einheitentypen",
        "__class__": None,
        "replace_column_names": None,
    },
}

# List of technologies which can be called by mastr.download()
#  as well as by MastrMirror.basic_backfill()
bulk_technologies = [
    "wind",
    "solar",
    "biomass",
    "hydro",
    "gsgk",
    "combustion",
    "nuclear",
    "gas",
    "storage",
    "electricity_consumer",
    "location",
    "market",
    "grid",
    "balancing_area",
    "permit",
]

# Map bulk technologies to bulk download tables
bulk_include_tables_map = {
    "wind": ["anlageneegwind", "einheitenwind"],
    "solar": ["anlageneegsolar", "einheitensolar"],
    "biomass": ["anlageneegbiomasse", "einheitenbiomasse"],
    "hydro": ["anlageneegwasser", "einheitenwasser"],
    "gsgk": [
        "anlageneeggeosolarthermiegrubenklaerschlammdruckentspannung",
        "einheitengeosolarthermiegrubenklaerschlammdruckentspannung",
    ],
    "combustion": ["anlagenkwk"],
    "nuclear": ["einheitenkernkraft"],
    "storage": ["anlageneegspeicher", "anlagenstromspeicher", "einheitenstromspeicher"],
    "gas": [
        "anlagengasspeicher",
        "einheitengaserzeuger",
        "einheitengasspeicher",
        "einheitengasverbraucher",
    ],
    "electricity_consumer": ["einheitenstromverbraucher", "einheitenverbrennung"],
    "location": ["lokationen"],
    "market": ["marktakteure", "marktrollen"],
    "grid": ["netzanschlusspunkte", "netze"],
    "balancing_area": ["bilanzierungsgebiete"],
    "permit": ["einheitengenehmigung"],
}



from open_mastr.utils.credentials import get_mastr_user, get_mastr_token


def test_get_mastr_user_and_token():

    user = get_mastr_user()
    assert len(user) == 15

    token = get_mastr_token(user)
    assert len(token) == 540



from datetime import date, datetime
from dateutil.parser import parse
import dateutil
import os
import sys
import shutil
import sqlite3
import open_mastr.settings as settings
from warnings import warn
import pandas as pd

# import xml dependencies
from os.path import expanduser
from open_mastr.xml_download.utils_download_bulk import download_xml_Mastr
from open_mastr.xml_download.utils_write_sqlite import convert_mastr_xml_to_sqlite

# import soap_API dependencies
from open_mastr.soap_api.mirror import MaStRMirror

# import initialize_database dependencies
from open_mastr.utils.helpers import db_engine
import open_mastr.orm as orm
from sqlalchemy.schema import CreateSchema


class Mastr:
    """
    :class:`.Mastr` is used to download the MaStR database and keep it up-to-date.

    A sqlite database is used to mirror the MaStR database. It can be filled with
    data either from the MaStR-bulk download or from the MaStR-API.

    .. code-block:: python

       from open_mastr import Mastr

       db = Mastr()
       db.download()
    """

    def __init__(self) -> None:

        # Define the paths for the zipped xml download and the sql databases
        self._xml_folder_path = os.path.join(
            expanduser("~"), ".open-MaStR", "data", "xml_download"
        )
        self._sqlite_folder_path = os.path.join(
            expanduser("~"), ".open-MaStR", "data", "sqlite"
        )
        os.makedirs(self._xml_folder_path, exist_ok=True)
        os.makedirs(self._sqlite_folder_path, exist_ok=True)

        # setup database engine and connection
        self.DB_ENGINE = os.environ.get("DB_ENGINE", "sqlite")
        self._engine = db_engine()
        SQLITE_DATABASE_PATH = settings.SQLITE_DATABASE_PATH
        self._sql_connection = sqlite3.connect(SQLITE_DATABASE_PATH)

        # Initialize database structure
        self._initialize_database()

    def download(
        self,
        method="bulk",
        technology=None,
        bulk_date_string="today",
        bulk_cleansing=True,
        api_processes=None,
        api_limit=50,
        api_date=None,
        api_chunksize=1000,
        api_data_types=None,
        api_location_types=None,
    ) -> None:
        """
        Download the MaStR either via the bulk download or via the MaStR API and write it to a
        sqlite database.

        Parameters
        ------------
        method: {'API', 'bulk'}, optional
            Either "API" or "bulk". Determines whether the data is downloaded via the
            zipped bulk download or via the MaStR API. The latter requires an account
            from marktstammdatenregister.de,
            (see :ref:`Configuration <Configuration>`). Default to 'bulk'.
        technology: str or list or None, optional
            Determines which technologies are written to the database. If None, all technologies are
            used. If it is a list, possible entries are "wind", "solar", "biomass", "hydro", "gsgk",
            "combustion", "nuclear", "gas", "storage", "electricity_consumer", "location", "market",
            "grid", "balancing_area" or "permit". If only one technology is of interest, this can be
            given as a string.  Default to None, where all technologies are included.
        bulk_date_string: str, optional
            Either "today" if the newest data dump should be downloaded from the MaStR website. If
            an already downloaded dump should be used, state the date of the download in the format
            "yyyymmdd". Default to "today".
        bulk_cleansing: bool, optional
            If True, data cleansing is applied after the download (which is recommended). Default
            to True.
        api_processes: int or None or "max", optional
            Number of parallel processes used to download additional data.
            Defaults to `None`. If set to "max", the maximum number of possible processes
            is used.
        api_limit: int or None, optional
            Limit number of units that data is download for. Defaults to `None` which refers
            to query data for existing data requests, for example created by
            :meth:`~.create_additional_data_requests`. Note: There is a limited number of
            requests you are allowed to have per day, so setting api_limit to a value is
            recommended.
        api_date: None or :class:`datetime.datetime` or str, optional
            Specify backfill date from which on data is retrieved

            Only data with modification time stamp greater that `date` is retrieved.

            * `datetime.datetime(2020, 11, 27)`: Retrieve data which is is newer than this
              time stamp
            * 'latest': Retrieve data which is newer than the newest data already in the table.

              .. warning::

                 Don't use "latest" in combination with "api_limit". This might lead to
                 unexpected results.

            * `None`: Complete backfill

            Defaults to `None`.
        api_chunksize: int or None, optional
            Data is downloaded and inserted into the database in chunks of `chunksize`.
            Defaults to 1000.
        api_data_types: list or None, optional
            Select type of additional data that should be retrieved. Choose from
            "unit_data", "eeg_data", "kwk_data", "permit_data".
        api_location_types: list or None, optional
            Select type of location that should be retrieved. Choose from
            "location_elec_generation", "location_elec_consumption", "location_gas_generation",
            "location_gas_consumption".
        """

        self._validate_parameter_format_for_download_method(
            method=method,
            technology=technology,
            bulk_date_string=bulk_date_string,
            bulk_cleansing=bulk_cleansing,
            api_processes=api_processes,
            api_limit=api_limit,
            api_date=api_date,
            api_chunksize=api_chunksize,
            api_data_types=api_data_types,
            api_location_types=api_location_types,
        )

        if method == "bulk":

            # Find the name of the zipped xml folder
            if bulk_date_string == "today":
                bulk_download_date = date.today().strftime("%Y%m%d")
            else:
                # proper format already tested in
                # self._validate_parameter_format_for_download_method
                bulk_download_date = parse(bulk_date_string).strftime("%Y%m%d")

            _zipped_xml_file_path = os.path.join(
                self._xml_folder_path,
                f"Gesamtdatenexport_{bulk_download_date}.zip",
            )

            if os.path.exists(_zipped_xml_file_path):
                print("MaStR already downloaded.")
            else:
                if bulk_date_string != "today":
                    raise Exception(
                        "There exists no file for given date. MaStR can only be downloaded "
                        "from the website if today's date is given."
                    )
                shutil.rmtree(self._xml_folder_path, ignore_errors=True)
                os.makedirs(self._xml_folder_path, exist_ok=True)
                download_xml_Mastr(_zipped_xml_file_path)
                print(f"MaStR was successfully downloaded to {self._xml_folder_path}.")

            # Map technology input to the .xml file names like 'einheitensolar'
            bulk_include_tables = self._technology_to_include_tables(
                technology=technology
            )

            convert_mastr_xml_to_sqlite(
                con=self._sql_connection,
                engine=self._engine,
                zipped_xml_file_path=_zipped_xml_file_path,
                include_tables=bulk_include_tables,
                bulk_cleansing=bulk_cleansing,
                bulk_download_date=bulk_download_date,
            )
            print("Bulk download and data cleansing was successful.")

        if method == "API":
            print(
                f"Downloading with soap_API.\n\n   -- Settings --  \nunits after date: "
                f"{api_date}\nunit download limit per technology: "
                f"{api_limit}\nparallel_processes: {api_processes}\nchunksize: "
                f"{api_chunksize}\ntechnologies: {technology}\ndata_types: "
                f"{api_data_types}\nlocation_types: {api_location_types}"
            )
            mastr_mirror = MaStRMirror(
                empty_schema=False,
                parallel_processes=api_processes,
                DB_ENGINE=self.DB_ENGINE,
                restore_dump=None,
            )
            # Download basic unit data
            mastr_mirror.backfill_basic(technology, limit=api_limit, date=api_date)

            # Download additional unit data
            for tech in technology:
                # mastr_mirror.create_additional_data_requests(tech)
                for data_type in api_data_types:
                    mastr_mirror.retrieve_additional_data(
                        tech, data_type, chunksize=api_chunksize, limit=api_limit
                    )

            # Download basic location data
            mastr_mirror.backfill_locations_basic(limit=api_limit, date="latest")

            # Download extended location data
            if api_location_types:
                for location_type in api_location_types:
                    mastr_mirror.retrieve_additional_location_data(
                        location_type, limit=api_limit
                    )

    def to_csv(self, path=None) -> None:
        """
        Save the database as csv files.

        Parameters
        ------------
        path: str or None
            The path where the csv files are saved.
            Default is None, in which case they are saved at
            HOME/.open-MaStR/data/csv
        """

        if path:
            if not os.path.exists(path):
                raise ValueError("parameter path is not a valid path")

        else:
            path = os.path.join(expanduser("~"), ".open-MaStR", "data", "csv")
        os.makedirs(path, exist_ok=True)

        engine = self._engine
        table_list = engine.table_names()
        with engine.connect().execution_options(autocommit=True) as con:
            print(f"csv files are saved to: {path}")
            for table in table_list:
                df = pd.read_sql(table, con=con)
                if not df.empty:
                    path_of_table = os.path.join(path, f"{table}.csv")
                    df.to_csv(path_or_buf=path_of_table, encoding="utf-16")
                    print(f"{table} saved as csv file.")

    def _initialize_database(self) -> None:
        engine = self._engine
        if self.DB_ENGINE == "docker":
            engine.execute(CreateSchema(orm.Base.metadata.schema))
        orm.Base.metadata.create_all(engine)

    def _validate_parameter_format_for_download_method(
        self,
        method,
        technology,
        bulk_date_string,
        bulk_cleansing,
        api_processes,
        api_limit,
        api_date,
        api_chunksize,
        api_data_types,
        api_location_types,
    ) -> None:
        # method parameter

        if method != "bulk" and method != "API":
            raise ValueError("parameter method has to be either 'bulk' or 'API'.")

        if method == "API":
            if bulk_cleansing is not True or bulk_date_string != "today":
                warn(
                    "For method = 'API', bulk download related parameters "
                    "(with prefix bulk_) are ignored."
                )

        if method == "bulk":
            if (
                any(
                    parameter is not None
                    for parameter in [
                        api_processes,
                        api_date,
                        api_data_types,
                        api_location_types,
                    ]
                )
                or api_limit != 50
                or api_chunksize != 1000
            ):
                warn(
                    "For method = 'bulk', API related parameters (with prefix api_) are ignored."
                )

        if not isinstance(technology, (str, list)) and technology is not None:
            raise ValueError("parameter technology has to be a string, list, or None")
        if isinstance(technology, str):
            technology = [technology]
        if isinstance(technology, list):
            bulk_technologies = [
                "wind",
                "solar",
                "biomass",
                "hydro",
                "gsgk",
                "combustion",
                "nuclear",
                "gas",
                "storage",
                "electricity_consumer",
                "location",
                "market",
                "grid",
                "balancing_area",
                "permit",
            ]
            for value in technology:
                if value not in bulk_technologies:
                    raise ValueError(
                        'Allowed values for parameter technology are "wind", "solar",'
                        '"biomass", "hydro", "gsgk", "combustion", "nuclear", "gas", '
                        '"storage", "electricity_consumer", "location", "market", '
                        '"grid", "balancing_area" or "permit"'
                    )

        if bulk_date_string != "today":
            try:
                _ = parse(bulk_date_string)
            except (dateutil.parser._parser.ParserError, TypeError):
                raise ValueError(
                    "parameter bulk_date_string has to be a proper date in the format yyyymmdd"
                    "or 'today'."
                )

        if type(bulk_cleansing) != bool:
            raise ValueError("parameter bulk_cleansing has to be boolean")

        if (
            api_processes != "max"
            and not isinstance(api_processes, int)
            and api_processes is not None
        ):
            raise ValueError(
                "parameter api_processes has to be 'max' or an integer or 'None'"
            )
        if api_processes == "max" or isinstance(api_processes, int):
            system = sys.platform
            if system not in ["linux2", "linux"]:
                raise ValueError(
                    "The functionality of multiprocessing only works on Linux based systems. "
                    "On your system, the parameter api_processes has to be 'None'."
                )

        if not isinstance(api_limit, int) and api_limit is not None:
            raise ValueError("parameter api_limit has to be an integer or 'None'.")

        if (
            not isinstance(api_date, datetime)
            and api_date != "latest"
            and api_date is not None
        ):
            raise ValueError(
                "parameter api_date has to be 'latest' or a datetime object or 'None'."
            )

        if not isinstance(api_chunksize, int) and api_chunksize is not None:
            raise ValueError("parameter api_chunksize has to be an integer or 'None'.")

        if not isinstance(api_data_types, list) and api_data_types is not None:
            raise ValueError("parameter api_data_types has to be a list or 'None'.")

        if isinstance(api_data_types, list):
            for value in api_data_types:
                if value not in [
                    "unit_data",
                    "eeg_data",
                    "kwk_data",
                    "permit_data",
                    None,
                ]:
                    raise ValueError(
                        'list entries of api_data_types have to be "unit_data", '
                        '"eeg_data", "kwk_data" '
                        'or "permit_data".'
                    )

        if not isinstance(api_location_types, list) and api_location_types is not None:
            raise ValueError("parameter api_location_types has to be a list or 'None'.")

        if isinstance(api_location_types, list):
            for value in api_location_types:
                if value not in [
                    "location_elec_generation",
                    "location_elec_consumption",
                    "location_gas_generation",
                    "location_gas_consumption",
                    None,
                ]:
                    raise ValueError(
                        'list entries of api_data_types have to be "location_elec_generation",'
                        '"location_elec_consumption", "location_gas_generation" or'
                        ' "location_gas_consumption".'
                    )

    def _technology_to_include_tables(
        self,
        technology,
        all_technologies=orm.bulk_technologies,
        tables_map=orm.bulk_include_tables_map,
    ) -> list:
        """
        Check the user input 'technology' and convert it to the list 'include_tables' which contains
        file names from zipped
        bulk download.
        Parameters
        ----------
        technology: None, str, list
            The user input for technology selection
            * `None`: All technologies (default)
            * `str`: One technology
            * `list`: List of technologies
        all_technologies: list
            All possible selections
        tables_map: dict
            Dictionary that maps the technologies to the file names in the zipped bulk
            download folder
        Returns
        -------
        list
            List of file names
        -------

        """
        # Convert technology input into a standard list
        chosen_technologies = []
        if technology is None:
            # All technologies are to be chosen
            chosen_technologies = all_technologies
        elif isinstance(technology, str):
            # Only one technology is chosen
            chosen_technologies = [technology]
        elif isinstance(technology, list):
            # list of technologies is given
            chosen_technologies = technology

        # Check if given technologies match with the valid options from 'orm.bulk_technologies'
        for tech in chosen_technologies:
            if tech not in all_technologies:
                raise Exception(
                    f"The input technology = {technology} does not match with the "
                    f"possible technology options. Only following technology options are available "
                    f"bulk_technologies = {all_technologies}"
                )

        # Map technologies to include tables
        include_tables = []
        for tech in chosen_technologies:
            # Append table names to the include_tables list respectively
            include_tables += tables_map[tech]

        return include_tables



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Service functions for OEP logging

Read data from MaStR API, process, and write to OEP

SPDX-License-Identifier: AGPL-3.0-or-later
"""

__copyright__ = "© Reiner Lemoine Institut"
__license__ = "GNU Affero General Public License Version 3 (AGPL-3.0)"
__url__ = "https://www.gnu.org/licenses/agpl-3.0.en.html"
__author__ = "Ludee; christian-rli"
__issue__ = "https://github.com/OpenEnergyPlatform/examples/issues/52"
__version__ = "v0.10.0"


import os
import open_mastr.soap_api.config as lc

from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, String, Float, DateTime, Boolean
from collections import namedtuple
import logging


log = logging.getLogger(__name__)
Base = declarative_base()

on_rtd = os.environ.get('READTHEDOCS', None) == 'True'


class Wind(Base):
    __tablename__ = 'mastr_wind'
    EinheitMastrNummer = Column(String(256))
    Id  = Column(Integer, primary_key=True)
    lid = Column(Integer)
    Name = Column(String(256))
    Einheitart = Column(String(256))
    Einheittyp = Column(String(256))
    Standort = Column(String(256))
    Bruttoleistung = Column(Float)
    Erzeugungsleistung = Column(Float)
    EinheitBetriebsstatus = Column(String(256))
    Anlagenbetreiber = Column(String(256))
    EegMastrNummer = Column(String(256))
    KwkMastrNummer = Column(String(256))
    SpeMastrNummer = Column(String(256))
    GenMastrNummer = Column(String(256))
    BestandsanlageMastrNummer = Column(String(256))
    NichtVorhandenInMigriertenEinheiten = Column(String(256))
    version = Column(String(256))
    timestamp = Column(DateTime)
    lid_w = Column(Integer)
    Ergebniscode = Column(String(256))
    AufrufVeraltet = Column(Boolean)
    AufrufLebenszeitEnde = Column(String(256))
    AufrufVersion = Column(String(256))
    DatumLetzteAktualisierung = Column(String(256))
    LokationMastrNummer = Column(String(256))
    NetzbetreiberpruefungStatus = Column(String(256))
    NetzbetreiberpruefungDatum = Column(String(256))
    NetzbetreiberMastrNummer = Column(String(256))
    Land = Column(String(256))
    Bundesland = Column(String(256))
    Landkreis = Column(String(256))
    Gemeinde = Column(String(256))
    Gemeindeschluessel = Column(String(256))
    Postleitzahl = Column(String(256))
    Gemarkung = Column(String(256))
    FlurFlurstuecknummern = Column(String(256))
    Strasse = Column(String(256))
    StrasseNichtGefunden = Column(Boolean)
    Hausnummer = Column(String(256))
    HausnummerNichtGefunden = Column(Boolean)
    Adresszusatz = Column(String(256))
    Ort = Column(String(256))
    Laengengrad = Column(String(256))
    Breitengrad = Column(String(256))
    UtmZonenwert = Column(String(256))
    UtmEast = Column(String(256))
    UtmNorth = Column(String(256))
    GaussKruegerHoch = Column(String(256))
    GaussKruegerRechts = Column(String(256))
    Meldedatum = Column(String(256))
    GeplantesInbetriebnahmedatum = Column(String(256))
    Inbetriebnahmedatum = Column(String(256))
    DatumEndgueltigeStilllegung = Column(String(256))
    DatumBeginnVoruebergehendeStilllegung = Column(String(256))
    DatumWiederaufnahmeBetrieb = Column(String(256))
    EinheitBetriebsstatus_w = Column(String(256))
    BestandsanlageMastrNummer_w = Column(String(256))
    NichtVorhandenInMigriertenEinheiten_w = Column(String(256))
    NameStromerzeugungseinheit = Column(String(256))
    Weic = Column(String(256))
    WeicDisplayName = Column(String(256))
    Kraftwerksnummer = Column(String(256))
    Energietraeger = Column(String(256))
    Bruttoleistung_w = Column(Float)
    Nettonennleistung = Column(String(256))
    AnschlussAnHoechstOderHochSpannung = Column(String(256))
    Schwarzstartfaehigkeit = Column(String(256))
    Inselbetriebsfaehigkeit = Column(String(256))
    Einsatzverantwortlicher = Column(String(256))
    FernsteuerbarkeitNb = Column(String(256))
    FernsteuerbarkeitDv = Column(String(256))
    FernsteuerbarkeitDr = Column(String(256))
    Einspeisungsart = Column(String(256))
    PraequalifiziertFuerRegelenergie = Column(String(256))
    GenMastrNummer_w = Column(String(256))
    NameWindpark = Column(String(256))
    Lage = Column(String(256))
    Seelage = Column(String(256))
    ClusterOstsee = Column(String(256))
    ClusterNordsee = Column(String(256))
    Technologie = Column(String(256))
    Typenbezeichnung = Column(String(256))
    Nabenhoehe = Column(String(256))
    Rotordurchmesser = Column(String(256))
    AuflageAbschaltungLeistungsbegrenzung = Column(String(256))
    Wassertiefe = Column(String(256))
    Kuestenentfernung = Column(String(256))
    EegMastrNummer_w = Column(String(256))
    HerstellerID = Column(String(256))
    HerstellerName = Column(String(256))
    version_w = Column(String(256))
    timestamp_w = Column(DateTime)
    lid_e = Column(Integer)
    Ergebniscode_e = Column(String(256))
    AufrufVeraltet_e = Column(Boolean)
    AufrufLebenszeitEnde_e = Column(String(256))
    AufrufVersion_e = Column(String(256))
    Meldedatum_e = Column(String(256))
    DatumLetzteAktualisierung_e = Column(String(256))
    EegInbetriebnahmedatum = Column(String(256))
    AnlagenkennzifferAnlagenregister = Column(String(256))
    AnlagenschluesselEeg = Column(String(256))
    PrototypAnlage = Column(String(256))
    PilotAnlage = Column(String(256))
    InstallierteLeistung = Column(String(256))
    VerhaeltnisReferenzertragErtrag5Jahre = Column(String(256))
    VerhaeltnisReferenzertragErtrag10Jahre = Column(String(256))
    VerhaeltnisReferenzertragErtrag15Jahre = Column(String(256))
    AusschreibungZuschlag = Column(String(256))
    Zuschlagsnummer = Column(String(256))
    AnlageBetriebsstatus = Column(String(256))
    VerknuepfteEinheit = Column(String(256))
    version_e = Column(String(256))
    timestamp_e = Column(String(256))


class Solar(Base):
    __tablename__ = 'mastr_solar'
    EinheitMastrNummer = Column(String(256))
    Id  = Column(Integer, primary_key=True)
    lid = Column(Integer)
    Name = Column(String(256))
    Einheitart = Column(String(256))
    Einheittyp = Column(String(256))
    Standort = Column(String(256))
    Bruttoleistung = Column(Float)
    Erzeugungsleistung = Column(Float)
    EinheitBetriebsstatus = Column(String(256))
    Anlagenbetreiber = Column(String(256))
    EegMastrNummer = Column(String(256))
    KwkMastrNummer = Column(String(256))
    SpeMastrNummer = Column(String(256))
    GenMastrNummer = Column(String(256))
    BestandsanlageMastrNummer = Column(String(256))
    NichtVorhandenInMigriertenEinheiten = Column(String(256))
    version = Column(String(256))
    timestamp = Column(DateTime)
    lid_w = Column(Integer)
    Ergebniscode = Column(String(256))
    AufrufVeraltet = Column(Boolean)
    AufrufLebenszeitEnde = Column(String(256))
    AufrufVersion = Column(String(256))
    DatumLetzteAktualisierung = Column(String(256))
    LokationMastrNummer = Column(String(256))
    NetzbetreiberpruefungStatus = Column(String(256))
    NetzbetreiberpruefungDatum = Column(String(256))
    NetzbetreiberMastrNummer = Column(String(256))
    Land = Column(String(256))
    Bundesland = Column(String(256))
    Landkreis = Column(String(256))
    Gemeinde = Column(String(256))
    Gemeindeschluessel = Column(String(256))
    Postleitzahl = Column(String(256))
    Gemarkung = Column(String(256))
    FlurFlurstuecknummern = Column(String(256))
    Strasse = Column(String(256))
    StrasseNichtGefunden = Column(Boolean)
    Hausnummer = Column(String(256))
    HausnummerNichtGefunden = Column(Boolean)
    Adresszusatz = Column(String(256))
    Ort = Column(String(256))
    Laengengrad = Column(String(256))
    Breitengrad = Column(String(256))
    UtmZonenwert = Column(String(256))
    UtmEast = Column(String(256))
    UtmNorth = Column(String(256))
    GaussKruegerHoch = Column(String(256))
    GaussKruegerRechts = Column(String(256))
    Meldedatum = Column(String(256))
    GeplantesInbetriebnahmedatum = Column(String(256))
    Inbetriebnahmedatum = Column(String(256))
    DatumEndgueltigeStilllegung = Column(String(256))
    DatumBeginnVoruebergehendeStilllegung = Column(String(256))
    DatumWiederaufnahmeBetrieb = Column(String(256))
    EinheitBetriebsstatus_w = Column(String(256))
    BestandsanlageMastrNummer_w = Column(String(256))
    NichtVorhandenInMigriertenEinheiten_w = Column(String(256))
    NameStromerzeugungseinheit = Column(String(256))
    Weic = Column(String(256))
    WeicDisplayName = Column(String(256))
    Kraftwerksnummer = Column(String(256))
    Energietraeger = Column(String(256))
    Bruttoleistung_w = Column(Float)
    Nettonennleistung = Column(String(256))
    AnschlussAnHoechstOderHochSpannung = Column(String(256))
    Schwarzstartfaehigkeit = Column(String(256))
    Inselbetriebsfaehigkeit = Column(String(256))
    Einsatzverantwortlicher = Column(String(256))
    FernsteuerbarkeitNb = Column(String(256))
    FernsteuerbarkeitDv = Column(String(256))
    FernsteuerbarkeitDr = Column(String(256))
    Einspeisungsart = Column(String(256))
    PraequalifiziertFuerRegelenergie = Column(String(256))
    GenMastrNummer_w = Column(String(256))
    NameWindpark = Column(String(256))
    Lage = Column(String(256))
    Seelage = Column(String(256))
    ClusterOstsee = Column(String(256))
    ClusterNordsee = Column(String(256))
    Technologie = Column(String(256))
    Typenbezeichnung = Column(String(256))
    Nabenhoehe = Column(String(256))
    Rotordurchmesser = Column(String(256))
    AuflageAbschaltungLeistungsbegrenzung = Column(String(256))
    Wassertiefe = Column(String(256))
    Kuestenentfernung = Column(String(256))
    EegMastrNummer_w = Column(String(256))
    HerstellerID = Column(String(256))
    HerstellerName = Column(String(256))
    version_w = Column(String(256))
    timestamp_w = Column(DateTime)
    lid_e = Column(Integer)
    Ergebniscode_e = Column(String(256))
    AufrufVeraltet_e = Column(Boolean)
    AufrufLebenszeitEnde_e = Column(String(256))
    AufrufVersion_e = Column(String(256))
    Meldedatum_e = Column(String(256))
    DatumLetzteAktualisierung_e = Column(String(256))
    EegInbetriebnahmedatum = Column(String(256))
    AnlagenkennzifferAnlagenregister = Column(String(256))
    AnlagenschluesselEeg = Column(String(256))
    PrototypAnlage = Column(String(256))
    PilotAnlage = Column(String(256))
    InstallierteLeistung = Column(String(256))
    VerhaeltnisReferenzertragErtrag5Jahre = Column(String(256))
    VerhaeltnisReferenzertragErtrag10Jahre = Column(String(256))
    VerhaeltnisReferenzertragErtrag15Jahre = Column(String(256))
    AusschreibungZuschlag = Column(String(256))
    Zuschlagsnummer = Column(String(256))
    AnlageBetriebsstatus = Column(String(256))
    VerknuepfteEinheit = Column(String(256))
    version_e = Column(String(256))
    timestamp_e = Column(String(256))

class Biomass(Base):
    __tablename__ = 'mastr_biomass'
    EinheitMastrNummer = Column(String(256))
    Id  = Column(Integer, primary_key=True)
    lid = Column(Integer)
    Name = Column(String(256))
    Einheitart = Column(String(256))
    Einheittyp = Column(String(256))
    Standort = Column(String(256))
    Bruttoleistung = Column(Float)
    Erzeugungsleistung = Column(Float)
    EinheitBetriebsstatus = Column(String(256))
    Anlagenbetreiber = Column(String(256))
    EegMastrNummer = Column(String(256))
    KwkMastrNummer = Column(String(256))
    SpeMastrNummer = Column(String(256))
    GenMastrNummer = Column(String(256))
    BestandsanlageMastrNummer = Column(String(256))
    NichtVorhandenInMigriertenEinheiten = Column(String(256))
    version = Column(String(256))
    timestamp = Column(DateTime)
    lid_w = Column(Integer)
    Ergebniscode = Column(String(256))
    AufrufVeraltet = Column(Boolean)
    AufrufLebenszeitEnde = Column(String(256))
    AufrufVersion = Column(String(256))
    DatumLetzteAktualisierung = Column(String(256))
    LokationMastrNummer = Column(String(256))
    NetzbetreiberpruefungStatus = Column(String(256))
    NetzbetreiberpruefungDatum = Column(String(256))
    NetzbetreiberMastrNummer = Column(String(256))
    Land = Column(String(256))
    Bundesland = Column(String(256))
    Landkreis = Column(String(256))
    Gemeinde = Column(String(256))
    Gemeindeschluessel = Column(String(256))
    Postleitzahl = Column(String(256))
    Gemarkung = Column(String(256))
    FlurFlurstuecknummern = Column(String(256))
    Strasse = Column(String(256))
    StrasseNichtGefunden = Column(Boolean)
    Hausnummer = Column(String(256))
    HausnummerNichtGefunden = Column(Boolean)
    Adresszusatz = Column(String(256))
    Ort = Column(String(256))
    Laengengrad = Column(String(256))
    Breitengrad = Column(String(256))
    UtmZonenwert = Column(String(256))
    UtmEast = Column(String(256))
    UtmNorth = Column(String(256))
    GaussKruegerHoch = Column(String(256))
    GaussKruegerRechts = Column(String(256))
    Meldedatum = Column(String(256))
    GeplantesInbetriebnahmedatum = Column(String(256))
    Inbetriebnahmedatum = Column(String(256))
    DatumEndgueltigeStilllegung = Column(String(256))
    DatumBeginnVoruebergehendeStilllegung = Column(String(256))
    DatumWiederaufnahmeBetrieb = Column(String(256))
    EinheitBetriebsstatus_w = Column(String(256))
    BestandsanlageMastrNummer_w = Column(String(256))
    NichtVorhandenInMigriertenEinheiten_w = Column(String(256))
    NameStromerzeugungseinheit = Column(String(256))
    Weic = Column(String(256))
    WeicDisplayName = Column(String(256))
    Kraftwerksnummer = Column(String(256))
    Energietraeger = Column(String(256))
    Bruttoleistung_w = Column(Float)
    Nettonennleistung = Column(String(256))
    AnschlussAnHoechstOderHochSpannung = Column(String(256))
    Schwarzstartfaehigkeit = Column(String(256))
    Inselbetriebsfaehigkeit = Column(String(256))
    Einsatzverantwortlicher = Column(String(256))
    FernsteuerbarkeitNb = Column(String(256))
    FernsteuerbarkeitDv = Column(String(256))
    FernsteuerbarkeitDr = Column(String(256))
    Einspeisungsart = Column(String(256))
    PraequalifiziertFuerRegelenergie = Column(String(256))
    GenMastrNummer_w = Column(String(256))
    NameWindpark = Column(String(256))
    Lage = Column(String(256))
    Seelage = Column(String(256))
    ClusterOstsee = Column(String(256))
    ClusterNordsee = Column(String(256))
    Technologie = Column(String(256))
    Typenbezeichnung = Column(String(256))
    Nabenhoehe = Column(String(256))
    Rotordurchmesser = Column(String(256))
    AuflageAbschaltungLeistungsbegrenzung = Column(String(256))
    Wassertiefe = Column(String(256))
    Kuestenentfernung = Column(String(256))
    EegMastrNummer_w = Column(String(256))
    HerstellerID = Column(String(256))
    HerstellerName = Column(String(256))
    version_w = Column(String(256))
    timestamp_w = Column(DateTime)
    lid_e = Column(Integer)
    Ergebniscode_e = Column(String(256))
    AufrufVeraltet_e = Column(Boolean)
    AufrufLebenszeitEnde_e = Column(String(256))
    AufrufVersion_e = Column(String(256))
    Meldedatum_e = Column(String(256))
    DatumLetzteAktualisierung_e = Column(String(256))
    EegInbetriebnahmedatum = Column(String(256))
    AnlagenkennzifferAnlagenregister = Column(String(256))
    AnlagenschluesselEeg = Column(String(256))
    PrototypAnlage = Column(String(256))
    PilotAnlage = Column(String(256))
    InstallierteLeistung = Column(String(256))
    VerhaeltnisReferenzertragErtrag5Jahre = Column(String(256))
    VerhaeltnisReferenzertragErtrag10Jahre = Column(String(256))
    VerhaeltnisReferenzertragErtrag15Jahre = Column(String(256))
    AusschreibungZuschlag = Column(String(256))
    Zuschlagsnummer = Column(String(256))
    AnlageBetriebsstatus = Column(String(256))
    VerknuepfteEinheit = Column(String(256))
    version_e = Column(String(256))
    timestamp_e = Column(String(256))

class Hydro(Base):
    __tablename__ = 'mastr_hydro'
    EinheitMastrNummer = Column(String(256))
    Id  = Column(Integer, primary_key=True)
    lid = Column(Integer)
    Name = Column(String(256))
    Einheitart = Column(String(256))
    Einheittyp = Column(String(256))
    Standort = Column(String(256))
    Bruttoleistung = Column(Float)
    Erzeugungsleistung = Column(Float)
    EinheitBetriebsstatus = Column(String(256))
    Anlagenbetreiber = Column(String(256))
    EegMastrNummer = Column(String(256))
    KwkMastrNummer = Column(String(256))
    SpeMastrNummer = Column(String(256))
    GenMastrNummer = Column(String(256))
    BestandsanlageMastrNummer = Column(String(256))
    NichtVorhandenInMigriertenEinheiten = Column(String(256))
    version = Column(String(256))
    timestamp = Column(DateTime)
    lid_w = Column(Integer)
    Ergebniscode = Column(String(256))
    AufrufVeraltet = Column(Boolean)
    AufrufLebenszeitEnde = Column(String(256))
    AufrufVersion = Column(String(256))
    DatumLetzteAktualisierung = Column(String(256))
    LokationMastrNummer = Column(String(256))
    NetzbetreiberpruefungStatus = Column(String(256))
    NetzbetreiberpruefungDatum = Column(String(256))
    NetzbetreiberMastrNummer = Column(String(256))
    Land = Column(String(256))
    Bundesland = Column(String(256))
    Landkreis = Column(String(256))
    Gemeinde = Column(String(256))
    Gemeindeschluessel = Column(String(256))
    Postleitzahl = Column(String(256))
    Gemarkung = Column(String(256))
    FlurFlurstuecknummern = Column(String(256))
    Strasse = Column(String(256))
    StrasseNichtGefunden = Column(Boolean)
    Hausnummer = Column(String(256))
    HausnummerNichtGefunden = Column(Boolean)
    Adresszusatz = Column(String(256))
    Ort = Column(String(256))
    Laengengrad = Column(String(256))
    Breitengrad = Column(String(256))
    UtmZonenwert = Column(String(256))
    UtmEast = Column(String(256))
    UtmNorth = Column(String(256))
    GaussKruegerHoch = Column(String(256))
    GaussKruegerRechts = Column(String(256))
    Meldedatum = Column(String(256))
    GeplantesInbetriebnahmedatum = Column(String(256))
    Inbetriebnahmedatum = Column(String(256))
    DatumEndgueltigeStilllegung = Column(String(256))
    DatumBeginnVoruebergehendeStilllegung = Column(String(256))
    DatumWiederaufnahmeBetrieb = Column(String(256))
    EinheitBetriebsstatus_w = Column(String(256))
    BestandsanlageMastrNummer_w = Column(String(256))
    NichtVorhandenInMigriertenEinheiten_w = Column(String(256))
    NameStromerzeugungseinheit = Column(String(256))
    Weic = Column(String(256))
    WeicDisplayName = Column(String(256))
    Kraftwerksnummer = Column(String(256))
    Energietraeger = Column(String(256))
    Bruttoleistung_w = Column(Float)
    Nettonennleistung = Column(String(256))
    AnschlussAnHoechstOderHochSpannung = Column(String(256))
    Schwarzstartfaehigkeit = Column(String(256))
    Inselbetriebsfaehigkeit = Column(String(256))
    Einsatzverantwortlicher = Column(String(256))
    FernsteuerbarkeitNb = Column(String(256))
    FernsteuerbarkeitDv = Column(String(256))
    FernsteuerbarkeitDr = Column(String(256))
    Einspeisungsart = Column(String(256))
    PraequalifiziertFuerRegelenergie = Column(String(256))
    GenMastrNummer_w = Column(String(256))
    NameWindpark = Column(String(256))
    Lage = Column(String(256))
    Seelage = Column(String(256))
    ClusterOstsee = Column(String(256))
    ClusterNordsee = Column(String(256))
    Technologie = Column(String(256))
    Typenbezeichnung = Column(String(256))
    Nabenhoehe = Column(String(256))
    Rotordurchmesser = Column(String(256))
    AuflageAbschaltungLeistungsbegrenzung = Column(String(256))
    Wassertiefe = Column(String(256))
    Kuestenentfernung = Column(String(256))
    EegMastrNummer_w = Column(String(256))
    HerstellerID = Column(String(256))
    HerstellerName = Column(String(256))
    version_w = Column(String(256))
    timestamp_w = Column(DateTime)
    lid_e = Column(Integer)
    Ergebniscode_e = Column(String(256))
    AufrufVeraltet_e = Column(Boolean)
    AufrufLebenszeitEnde_e = Column(String(256))
    AufrufVersion_e = Column(String(256))
    Meldedatum_e = Column(String(256))
    DatumLetzteAktualisierung_e = Column(String(256))
    EegInbetriebnahmedatum = Column(String(256))
    AnlagenkennzifferAnlagenregister = Column(String(256))
    AnlagenschluesselEeg = Column(String(256))
    PrototypAnlage = Column(String(256))
    PilotAnlage = Column(String(256))
    InstallierteLeistung = Column(String(256))
    VerhaeltnisReferenzertragErtrag5Jahre = Column(String(256))
    VerhaeltnisReferenzertragErtrag10Jahre = Column(String(256))
    VerhaeltnisReferenzertragErtrag15Jahre = Column(String(256))
    AusschreibungZuschlag = Column(String(256))
    Zuschlagsnummer = Column(String(256))
    AnlageBetriebsstatus = Column(String(256))
    VerknuepfteEinheit = Column(String(256))
    version_e = Column(String(256))
    timestamp_e = Column(String(256))



from shutil import Error
from zipfile import ZipFile
import lxml
import numpy as np
import pandas as pd
import sqlalchemy
import sqlite3
from open_mastr.orm import tablename_mapping
from open_mastr.xml_download.colums_to_replace import system_catalog
from open_mastr.xml_download.utils_sqlite_bulk_cleansing import (
    replace_mastr_katalogeintraege,
    date_columns_to_datetime,
)
from sqlalchemy import (
    Integer,
    String,
    Float,
    DateTime,
    Boolean,
    Date,
    JSON,
)


dtypes_mapping = {
    String: "string",
    Integer: "int",
    Float: "float",
    DateTime: "datetime64[s]",
    Boolean: "bool",
    Date: "datetime64[s]",
    JSON: "string",
}


def convert_mastr_xml_to_sqlite(
    con: sqlite3.Connection,
    engine,
    zipped_xml_file_path: str,
    include_tables: list,
    bulk_cleansing: bool,
    bulk_download_date: str,
) -> None:
    """Converts the Mastr in xml format into a sqlite database."""
    """Writes the local zipped MaStR to a PostgreSQL database.

    Parameters
    ------------
    include_tables : list, default None
        List of tables from the Marktstammdatenregister that
        should be written into the database. Elements of
        include_tables are lower case strings without "_" and index.
        It is possible to include any table from the zipped
        local MaStR folder (see MaStR.initialize()).
        Example: If you do want to write the data from
        files "AnlagenEegSolar_*.xml" to a table in your
        database (where * is any number >=1), write the
        element "anlageneegsolar" into the include_tables list.
        Elements of the list that cannot be matched to tables of
        the MaStR are ignored. If include_tables is given,
        only the tables listed here are written to the database.


    """

    with ZipFile(zipped_xml_file_path, "r") as f:
        for file_name in f.namelist():
            # xml_tablename is the beginning of the filename without the number in lowercase
            xml_tablename = file_name.split("_")[0].split(".")[0].lower()

            # few tables are only needed for data cleansing of the xml files and contain no
            # information of relevance
            boolean_write_table_to_sql_database = False
            if tablename_mapping[xml_tablename]["__class__"] is not None:
                boolean_write_table_to_sql_database = True

            # check if the table should be written to sql database (depends on user input)
            include_count = include_tables.count(xml_tablename)

            if include_count == 1 and boolean_write_table_to_sql_database:

                sql_tablename = tablename_mapping[xml_tablename]["__name__"]
                # check if the file name indicates that it is the first file from the
                # table
                if (
                    file_name.split(".")[0].split("_")[-1] == "1"
                    or len(file_name.split(".")[0].split("_")) == 1
                ):
                    orm_class = tablename_mapping[xml_tablename]["__class__"]
                    # drop the content from table
                    orm_class.__table__.drop(engine, checkfirst=True)
                    # create table schema
                    orm_class.__table__.create(engine)
                    index_for_printed_message = 1
                    print(
                        f"Table '{sql_tablename}' is filled with data '{xml_tablename}' "
                        "from the bulk download. \n"
                        f"File '{file_name}' is parsed."
                    )
                    index_for_printed_message += 1
                else:
                    print(
                        f"File '{file_name}' is parsed."
                    )
                    index_for_printed_message += 1

                df = prepare_table_to_sqlite_database(
                    f=f,
                    file_name=file_name,
                    xml_tablename=xml_tablename,
                    bulk_download_date=bulk_download_date,
                )

                # Convert date and datetime columns into the datatype datetime
                df = date_columns_to_datetime(xml_tablename, df)

                if bulk_cleansing:
                    print("Data is cleansed.")
                    # Katalogeintraege: int -> string value
                    df = replace_mastr_katalogeintraege(
                        zipped_xml_file_path=zipped_xml_file_path, df=df
                    )

                add_table_to_sqlite_database(
                    df=df,
                    xml_tablename=xml_tablename,
                    sql_tablename=sql_tablename,
                    if_exists="append",
                    con=con,
                    engine=engine,
                )


def prepare_table_to_sqlite_database(
    f: ZipFile,
    file_name: str,
    xml_tablename: str,
    bulk_download_date: str,
) -> pd.DataFrame:
    data = f.read(file_name)
    try:
        df = pd.read_xml(data, encoding="UTF-16", compression="zip")
    except lxml.etree.XMLSyntaxError as err:
        df = handle_xml_syntax_error(data, err)

    # Change data types in dataframe
    dict_of_columns_and_string_length = {
        "Gemeindeschluessel": 8,
        "Postleitzahl": 5,
    }
    for column_name in dict_of_columns_and_string_length.keys():
        if column_name in df.columns:
            string_length = dict_of_columns_and_string_length[column_name]
            df = add_zero_as_first_character_for_too_short_string(
                df, column_name, string_length
            )

    # Replace IDs with names from system_catalogue
    for column_name in system_catalog.keys():
        if column_name in df.columns:
            df[column_name] = df[column_name].replace(system_catalog[column_name])

    # Change column names according to orm data model
    if tablename_mapping[xml_tablename]["replace_column_names"]:

        df = df.rename(columns=tablename_mapping[xml_tablename]["replace_column_names"])

    # Add Column that refers to the source of the data
    df["DatenQuelle"] = "bulk"
    df["DatumDownload"] = bulk_download_date
    return df


def add_table_to_sqlite_database(
    df: pd.DataFrame,
    xml_tablename: str,
    sql_tablename: str,
    if_exists: str,
    con: sqlite3.Connection,
    engine,
) -> None:

    # get a dictionary for the data types
    dtypes_for_writing_sql = {}

    for column in list(tablename_mapping[xml_tablename]["__class__"].__table__.columns):
        if column.name in df.columns:
            dtypes_for_writing_sql[column.name] = column.type

    continueloop = True
    while continueloop:
        try:
            df.to_sql(
                sql_tablename,
                con=engine,
                index=False,
                if_exists=if_exists,
                dtype=dtypes_for_writing_sql,
            )
            continueloop = False
        except sqlalchemy.exc.OperationalError as err:
            add_missing_column_to_table(err, con, sql_tablename)

        except sqlite3.OperationalError as err:
            add_missing_column_to_table(err, con, sql_tablename)

        except sqlalchemy.exc.DataError as err:
            delete_wrong_xml_entry(err, df)

        except sqlalchemy.exc.IntegrityError as err:
            # error resulting from Unique constraint failed
            df = write_single_entries_until_not_unique_comes_up(
                df=df, sql_tablename=sql_tablename, engine=engine, err=err
            )


def add_zero_as_first_character_for_too_short_string(df, column_name, string_length):
    # Gemeindeschluessel or PLZ are read as float, but they are actually strings
    # if they start with a 0 this gets lost
    try:
        df[column_name] = df[column_name].astype("Int64").astype(str)
    except ValueError:
        # some Plz are in the format DK-9999 for danish Postleitzahl
        # They cannot be converted to integer
        df[column_name] = df[column_name].astype(str)

    df[column_name] = df[column_name].where(cond=df[column_name] != "None", other=None)
    df[column_name] = df[column_name].where(cond=df[column_name] != "<NA>", other=None)

    string_adding_series = pd.Series(["0"] * len(df))
    string_adding_series = string_adding_series.where(
        cond=df[column_name].str.len() == string_length - 1, other=""
    )
    df[column_name] = string_adding_series + df[column_name]
    return df


def write_single_entries_until_not_unique_comes_up(
    df: pd.DataFrame, sql_tablename: str, engine: sqlalchemy.engine.Engine, err: str
) -> None:

    key_column = str(err).split("\n[SQL: INSERT INTO")[0].split("UNIQUE constraint failed:")[1].split(".")[1]
    key_list = pd.read_sql(sql = f"SELECT {key_column} FROM {sql_tablename};", con=engine).values.squeeze().tolist()
    df = df.set_index(key_column)
    len_df_before = len(df)
    df = df.drop(labels=key_list, errors='ignore')
    df = df.reset_index()
    print(f"{len_df_before-len(df)} entries already existed in the database.")

    return df


def add_missing_column_to_table(
    err: Error, con: sqlite3.Connection, sql_tablename: str
) -> None:
    """Some files introduce new columns for existing tables.
    If this happens, the error from writing entries into
    non-existing columns is caught and the column is created."""

    # Needed for sqlite3 error message
    # missing_column = str(err).split("no column named ")[1]

    # Needed for sqlalchemy error message
    missing_column = str(err).split("has no column named ")[1].split("\n[SQL: ")[0]

    cursor = con.cursor()
    execute_message = 'ALTER TABLE %s ADD "%s" VARCHAR NULL;' % (
        sql_tablename,
        missing_column,
    )
    cursor.execute(execute_message)
    con.commit()
    cursor.close()


def delete_wrong_xml_entry(err: Error, df: pd.DataFrame) -> None:
    delete_entry = str(err).split("«")[0].split("»")[1]
    print(f"The entry {delete_entry} was deleteted due to its false data type.")
    df = df.replace(delete_entry, np.nan)


def handle_xml_syntax_error(data: bytes, err: Error) -> pd.DataFrame:
    """Deletes entries that cause an xml syntax error and produces DataFrame.

    Parameters
    -----------
    data : bytes
        Unzipped xml data
    err : ErrorMessage
        Error message that appeared when trying to use pd.read_xml on invalid xml file.

    Returns
    ----------
    df : pandas.DataFrame
        DataFrame which is read from the changed xml data.
    """
    wrong_char_position = int(str(err).split()[-4])
    decoded_data = data.decode("utf-16")
    loop_condition = True

    shift = 0
    while loop_condition:
        evaluated_string = decoded_data[wrong_char_position + shift]
        if evaluated_string == ">":
            start_char = wrong_char_position + shift + 1
            break
        else:
            shift -= 1
    loop_condition_2 = True
    while loop_condition_2:
        evaluated_string = decoded_data[start_char]
        if evaluated_string == "<":
            break
        else:
            decoded_data = decoded_data[:start_char] + decoded_data[start_char + 1 :]
    df = pd.read_xml(decoded_data)
    print("One invalid xml expression was deleted.")
    return df



from io import BytesIO
import re
from urllib.request import urlopen
from zipfile import ZipFile
import xmltodict
from collections import OrderedDict


class DataDescription(object):
    """
    Fundamental data for creating a description of open-MaStR data

    Extract documentation data from MaStR webservice XML file.

    .. code:: python

       functions_data_docs = DataDescription().functions_data_documentation()
    """

    def __init__(self, xml=None):
        """

        Parameters
        ----------
        xml: str or path-like, optional
            Path of local mastrbasetypes.xsd file. If not provided, file will
            be downloaded from marktstammdatenregister.de
        """

        # Read XML file
        if xml:
            with open(xml, "r") as fh:
                self.xml = fh.read()
        else:
            # If no XML file is given, the file is read from an URL
            zipurl = 'https://www.marktstammdatenregister.de/MaStRHilfe/files/' \
                     'webdienst/Dienstbeschreibung_1_2_39_Produktion.zip'

            with urlopen(zipurl) as zipresp:
                with ZipFile(BytesIO(zipresp.read())) as zfile:
                    self.xml = zfile.read('xsd/mastrbasetypes.xsd')



        # Parse XML and extract relevant data
        parsed = xmltodict.parse(self.xml, process_namespaces=False)
        self.complex_types = parsed['schema']["complexType"]
        self.simple_types = parsed['schema']["simpleType"]

        # Prepare parsed data for documentational purposes
        abstract_types, parameters, responses, types = self._filter_type_descriptions()
        self.function_parameters = parameters
        self.function_responses = responses
        self.types = {**abstract_types, **types}
        self.simple_types_prepared = self.prepare_simple_type()

    def _filter_type_descriptions(self):
        """
        Reorganize complex data types
        """

        functions = []
        parameters = {}
        responses = {}
        abstract_types = {}
        types = {}

        for item in self.complex_types:

            # Filter out abstract types
            # Those are use to inherit columns from
            if "Basis" in item["@name"]:
                if "sequence" in item.keys():
                    abstract_types[item["@name"]] = item["sequence"]
                elif "complexContent" in item.keys():
                    abstract_types[item["@name"]] = item["complexContent"]
                else:
                    raise ValueError("Ohh...")
            else:
                # Filter all functions
                if item["@name"].startswith(("Get", "Set", "Erneute", "Verschiebe", "Delete")):
                    functions.append(item)

                    # Further split the list of functions into paramters and responses
                    if item["@name"].endswith("Parameter"):
                        if "complexContent" in item.keys():
                            parameters[item["@name"]] = item["complexContent"]["extension"]
                        else:
                            parameters[item["@name"]] = item
                    elif item["@name"].endswith("Antwort"):
                        # responses.append(item)
                        responses[item["@name"]] = item["complexContent"]["extension"]
                    else:
                        raise ValueError("Should be Parameter or Antwort!")
                else:
                    types[item["@name"]] = item

        return abstract_types, parameters, responses, types

    def prepare_simple_type(self):
        """
        Reformatting and filtering of simple MaStR type documentation

        Returns
        -------
        dict
            Data type and possible values for each simple MaStR type keyed by type name
        """

        simple_types_doc = {}

        for simple_type in self.simple_types:
            if "enumeration" in simple_type["restriction"]:
                possible_values = [_["@value"] for _ in simple_type["restriction"]["enumeration"]]
            else:
                possible_values = []
            simple_types_doc[simple_type["@name"]] = {
                "type": simple_type["restriction"]["@base"],
                "values": possible_values
            }
        return simple_types_doc

    def functions_data_documentation(self):
        """
        Documentation for the data returned by MaStR webservice functions

        Returns
        -------
        dict
            Documentation of each column/field returned by a function keyed by function names
        """
        function_docs = {}
        for name, fcn in self.function_responses.items():
            fcn_name = name.replace("Antwort", "")
            if fcn["sequence"]:
                # Slice data depending on what is available
                if isinstance(fcn["sequence"]["element"], list):
                    fcn_data = fcn["sequence"]["element"]
                elif isinstance(fcn["sequence"]["element"], (dict, OrderedDict)):
                    if "annotation" in fcn["sequence"]["element"]:
                        fcn_data = [fcn["sequence"]["element"]]
                    else:
                        fcn_data = self.types[fcn["sequence"]["element"]["@type"].split(":")[1]]["sequence"]["element"]
                else:
                    print(type(fcn["sequence"]))
                    print(fcn["sequence"])
                    raise ValueError

                # Add data for inherited columns from base types
                if "@base" in fcn:
                    if not fcn["@base"] == 'mastr:AntwortBasis':
                        fcn_data = _collect_columns_of_base_type(self.types, fcn["@base"].split(":")[1], fcn_data)
                function_docs[fcn_name] = {}
                for column in fcn_data:
                    # Replace MaStR internal types with more general ones
                    if column["@type"].startswith("mastr:"):
                        try:
                            column_type = self.simple_types_prepared[column["@type"].split(":")[1]]["type"]
                        except KeyError:
                            column_type = column["@type"]
                    else:
                        column_type = column["@type"]

                    if "annotation" in column.keys():
                        description = column["annotation"]["documentation"].get("#text", None)
                        if description:
                            description = re.sub(" +", " ", description.replace("\n", ""))
                        function_docs[fcn_name][column["@name"]] = {
                                "type":  column_type,
                                "description": description,
                                "example": column["annotation"]["documentation"].get("m-ex", None)
                        }
                    else:
                        function_docs[fcn_name][column["@name"]] = {
                            "type": column_type,
                            # TODO: insert information from simple type here
                            "description": None,
                            "example": None
                        }

        # Hack in a descrition for a column that gets created after download while flattening data
        function_docs["GetEinheitWind"]["HerstellerId"] = {
            "type": "str",
            "description": "Id des Herstellers der Einheit",
            "example": 923
        }

        return function_docs


def _collect_columns_of_base_type(base_types, base_type_name, fcn_data):
    type_description = base_types[base_type_name]
    fcn_data += type_description["extension"]["sequence"]["element"]

    if "@base" in type_description["extension"]:
        if not type_description["extension"]["@base"] == 'mastr:AntwortBasis':
            fcn_data = _collect_columns_of_base_type(base_types, type_description["extension"]["@base"].split(":")[1], fcn_data)

    return fcn_data



import pandas as pd
import numpy as np
import sqlalchemy

from open_mastr.xml_download.colums_to_replace import columns_replace_list
from zipfile import ZipFile
from open_mastr.orm import tablename_mapping


def replace_mastr_katalogeintraege(
    zipped_xml_file_path: str,
    df: pd.DataFrame,
) -> pd.DataFrame:

    katalogwerte = create_katalogwerte_from_sqlite(zipped_xml_file_path)

    for column_name in df.columns:
        if column_name in columns_replace_list:
            df[column_name] = (
                df[column_name].astype("float").astype("Int64").map(katalogwerte)
            )

    return df


def create_katalogwerte_from_sqlite(zipped_xml_file_path) -> dict:
    with ZipFile(zipped_xml_file_path, "r") as f:
        data = f.read("Katalogwerte.xml")
        df_katalogwerte = pd.read_xml(data, encoding="UTF-16", compression="zip")
    katalogwerte_array = np.array(df_katalogwerte[["Id", "Wert"]])
    katalogwerte = dict(
        (katalogwerte_array[n][0], katalogwerte_array[n][1])
        for n in range(len(katalogwerte_array))
    )
    return katalogwerte


def date_columns_to_datetime(xml_tablename: str, df: pd.DataFrame) -> pd.DataFrame:

    sqlalchemy_columnlist = tablename_mapping[xml_tablename][
        "__class__"
    ].__table__.columns.items()

    # Iterate over all columns from the orm data class
    # If the column has data type = date or datetime, the function
    # pd.to_datetime is applied
    for column in sqlalchemy_columnlist:
        if (
            type(column[1].type) == sqlalchemy.sql.sqltypes.Date
            or type(column[1].type) == sqlalchemy.sql.sqltypes.DateTime
        ):
            column_name = column[0]
            if column_name in df.columns:
                # Convert column to datetime64, invalid string -> NaT
                df[column_name] = pd.to_datetime(df[column_name], errors="coerce")
    return df



# __copyright__ = "© Reiner Lemoine Institut"
# __license__ = "GNU Affero General Public License Version 3 (AGPL-3.0)"
# __url__ = "https://www.gnu.org/licenses/agpl-3.0.en.html"
# __author__ = "Bachibouzouk"
# __version__ = "v0.10.0"
#
# import time
# import pandas as pd
# import os
#
# from soap_api.sessions import API_MAX_DEMANDS
#
# from soap_api.mastr_power_unit_download import (
#     get_power_unit,
#     download_power_unit,
#     download_parallel_power_unit
# )
#
# TEST_DATA = 'tests/data/bnetza_mastr_test_power-unit.csv'
#
#
# def remove_test_data(fname=TEST_DATA):
#     time.sleep(0.1)
#     os.remove(fname)
#     time.sleep(0.1)
#     try:
#         os.rmdir(os.path.dirname(fname))
#     except OSError:
#         print('could not delete {}'.format(os.path.dirname(fname)))
#
#
# def test_get_power_unit_return_data_frame():
#     power_units = get_power_unit(0, 10)
#     assert isinstance(power_units, pd.DataFrame)
#
#
# def test_get_power_unit_return_correct_number():
#     n = 13
#     power_units = get_power_unit(0, n)
#     assert len(power_units.index) == n
#
#
# def test_get_power_unit_starting_not_from_zero():
#     n = 13
#     power_units = get_power_unit(1000, n)
#     assert len(power_units.index) == n
#
#
# def test_get_power_unit_largest_limit():
#     power_units = get_power_unit(0, API_MAX_DEMANDS)
#     assert len(power_units.index) == API_MAX_DEMANDS
#
#
# def test_download_power_unit_unique_mastr_number():
#     n = 100
#     # download the data
#     download_power_unit(n, ofname=TEST_DATA)
#     # load the data
#     df = pd.read_csv(TEST_DATA, sep=';')
#     # remove the data
#     remove_test_data()
#     assert len(df.EinheitMastrNummer.unique()) == n
#     assert len(df.EegMastrNummer.unique()) == n
#
#
# def test_parallel_download_power_unit_unique_mastr_number_case1():
#     """Number is exactly the same as the API_MAX_DEMAND"""
#     n = API_MAX_DEMANDS
#     # download the data
#     download_parallel_power_unit(n, ofname=TEST_DATA)
#     # load the data
#     df = pd.read_csv(TEST_DATA, sep=';')
#     # remove the data
#     remove_test_data()
#     assert len(df.EinheitMastrNummer.unique()) == n
#     assert len(df.EegMastrNummer.unique()) == n
#
#
# def test_parallel_download_power_unit_unique_mastr_number_case2():
#     """Number is an integer number of the API_MAX_DEMAND"""
#
#     n = 2 * API_MAX_DEMANDS
#     # download the data
#     download_parallel_power_unit(n, ofname=TEST_DATA)
#     # load the data
#     df = pd.read_csv(TEST_DATA, sep=';')
#     # remove the data
#     remove_test_data()
#     assert len(df.EinheitMastrNummer.unique()) == n
#     assert len(df.EegMastrNummer.unique()) == n
#
#
# def test_parallel_download_power_unit_unique_mastr_number_case3():
#     """Number is larger than the API_MAX_DEMAND"""
#
#     n = int(1.5 * API_MAX_DEMANDS)
#     # download the data
#     download_parallel_power_unit(n, ofname=TEST_DATA)
#     # load the data
#     df = pd.read_csv(TEST_DATA, sep=';')
#     # remove the data
#     remove_test_data()
#     assert len(df.EinheitMastrNummer.unique()) == n
#     assert len(df.EegMastrNummer.unique()) == n
#
#
# def test_parallel_download_power_unit_unique_mastr_number_case4():
#     """Number is larger than the API_MAX_DEMAND and batch_size"""
#
#     n = 6 * API_MAX_DEMANDS
#     # download the data
#     download_parallel_power_unit(n, ofname=TEST_DATA)
#     # load the data
#     df = pd.read_csv(TEST_DATA, sep=';')
#     # remove the data
#     remove_test_data()
#     assert len(df.EinheitMastrNummer.unique()) == n
#     assert len(df.EegMastrNummer.unique()) == n